import sys
from pytest import mark, approx

@mark.skipif(sys.platform != "linux", reason="Only works under linux")
def test_calculate_rat_rcwa():
    import numpy as np

    from solcore import material, si
    from solcore.structure import Layer, Structure
    from solcore.absorption_calculator.rigorous_coupled_wave import calculate_rat_rcwa

    T = 300
    Air = material("Air")(T=T)
    TiO2 = material("TiO2", sopra=True)(T=T)  # for the nanoparticles
    GaAs = material("GaAs")(T=T)

    NP_layer = Layer(
        si("50nm"),
        Air,
        geometry=[{"type": "circle", "mat": TiO2, "center": (200, 200), "radius": 50}],
    )
    np_struct = Structure([NP_layer])
    wl = np.linspace(300, 1000, 150)

    rat_np = calculate_rat_rcwa(
        np_struct, size=((400, 0), (0, 400)), orders=10, wavelength=wl, substrate=GaAs, incidence=Air
    )

    assert sorted(list(rat_np.keys())) == ["A", "A_per_layer", "A_pol", "R", "T"]
    for v in rat_np.values():
        assert v.shape[0] == len(wl)


@mark.skipif(sys.platform != "linux", reason="Only works under linux")
def test_calculate_absorption_profile_rcwa():
    import numpy as np

    from solcore import material, si
    from solcore.structure import Layer, Structure
    from solcore.absorption_calculator.rigorous_coupled_wave import calculate_rat_rcwa, calculate_absorption_profile_rcwa

    T = 300
    Air = material("Air")(T=T)
    TiO2 = material("TiO2", sopra=True)(T=T)  # for the nanoparticles
    GaAs = material("GaAs")(T=T)
    th = 50
    NP_layer = Layer(
        si(th, "nm"),
        Air,
        geometry=[{"type": "ellipse", "mat": TiO2, "center": (200, 200), "halfwidths": [70, 100], "angle": 30}],
    )
    np_struct = Structure([NP_layer])
    wl = np.linspace(300, 1000, 10)

    rat_np = calculate_rat_rcwa(
        np_struct, size=((400, 0), (0, 400)), orders=10, wavelength=wl, substrate=GaAs, incidence=Air
    )

    A_output = rat_np['A_pol']
    step_size = 2
    dist = np.arange(0, th, step_size)
    result = calculate_absorption_profile_rcwa(np_struct, size=((400, 0), (0, 400)), orders=10, wavelength=wl,
                                               rat_output_A=A_output, steps_size=step_size)

    parallel_result = calculate_absorption_profile_rcwa(np_struct, size=((400, 0), (0, 400)), orders=10, wavelength=wl,
                                               rat_output_A=A_output, parallel=True, steps_size=step_size)

    assert sorted(list(result.keys())) == ["absorption", "position"]
    assert sorted(list(parallel_result.keys())) == ["absorption", "position"]

    assert len(result["position"]) == len(dist)
    assert len(parallel_result["position"]) == len(dist)

    assert approx(result["absorption"] == parallel_result["absorption"])


@mark.skipif(sys.platform != "linux", reason="Only works under linux")
def test_pol_rcwa():
    import numpy as np

    from solcore import material, si
    from solcore.structure import Layer, Structure
    from solcore.absorption_calculator.rigorous_coupled_wave import calculate_rat_rcwa, \
        calculate_absorption_profile_rcwa

    T = 300
    Air = material("Air")(T=T)
    TiO2 = material("TiO2", sopra=True)(T=T)  # for the nanoparticles
    GaAs = material("GaAs")(T=T)
    th = 50
    NP_layer = Layer(
        si(th, "nm"),
        Air,
        geometry=[{"type": "rectangle", "mat": TiO2, "center": (200, 200), "halfwidths": [100, 70], "angle": 40}],
    )
    np_struct = Structure([NP_layer])
    wl = np.linspace(300, 1000, 10)

    step_size = 2

    rat_np_u = calculate_rat_rcwa(
        np_struct, size=((400, 0), (0, 400)), orders=10, wavelength=wl, substrate=GaAs, incidence=Air, pol='u'
    )

    A_output_u = rat_np_u['A_pol']
    result_u = calculate_absorption_profile_rcwa(np_struct, size=((400, 0), (0, 400)), orders=10, wavelength=wl,
                                                 rat_output_A=A_output_u, parallel=True, steps_size=step_size,
                                                 pol='u')

    rat_np_s = calculate_rat_rcwa(
        np_struct, size=((400, 0), (0, 400)), orders=10, wavelength=wl, substrate=GaAs, incidence=Air, pol='s'
    )

    A_output_s = rat_np_s['A_pol']
    result_s = calculate_absorption_profile_rcwa(np_struct, size=((400, 0), (0, 400)), orders=10, wavelength=wl,
                                                 rat_output_A=A_output_s, parallel=True, steps_size=step_size,
                                                 pol='s')

    rat_np_p = calculate_rat_rcwa(
        np_struct, size=((400, 0), (0, 400)), orders=10, wavelength=wl, substrate=GaAs, incidence=Air, pol='p'
    )

    A_output_p = rat_np_p['A_pol']
    result_p = calculate_absorption_profile_rcwa(np_struct, size=((400, 0), (0, 400)), orders=10, wavelength=wl,
                                                 rat_output_A=A_output_p, parallel=True, steps_size=step_size,
                                                 pol='p')

    assert approx(A_output_u == 0.5*(A_output_p + A_output_s))

    assert approx(result_u["absorption"] == 0.5*(result_s["absorption"] + result_p["absorption"]))


@mark.skipif(sys.platform != "linux", reason="Only works under linux")
def test_arbitrary_pol_rcwa():
    import numpy as np

    from solcore import material, si
    from solcore.structure import Layer, Structure
    from solcore.absorption_calculator.rigorous_coupled_wave import calculate_rat_rcwa, \
        calculate_absorption_profile_rcwa

    T = 300
    Air = material("Air")(T=T)
    TiO2 = material("TiO2", sopra=True)(T=T)  # for the nanoparticles
    GaAs = material("GaAs")(T=T)
    th = 50
    NP_layer = Layer(
        si(th, "nm"),
        Air,
        geometry=[{"type": "ellipse", "mat": TiO2, "center": (200, 200), "halfwidths": [100, 70], "angle": 40}],
    )
    np_struct = Structure([NP_layer])
    wl = np.linspace(300, 1000, 10)

    step_size = 2

    rat_np = calculate_rat_rcwa(
        np_struct, size=((400, 0), (0, 400)), orders=10, wavelength=wl, substrate=GaAs, incidence=Air,
        pol=[1, 0]
    )

    A_output = rat_np['A_pol']
    result = calculate_absorption_profile_rcwa(np_struct, size=((400, 0), (0, 400)), orders=10, wavelength=wl,
                                                 rat_output_A=A_output, parallel=True, steps_size=step_size,
                                                 pol=[1, 0])

    rat_np_s = calculate_rat_rcwa(
        np_struct, size=((400, 0), (0, 400)), orders=10, wavelength=wl, substrate=GaAs, incidence=Air, pol='s'
    )

    A_output_s = rat_np_s['A_pol']
    result_s = calculate_absorption_profile_rcwa(np_struct, size=((400, 0), (0, 400)), orders=10, wavelength=wl,
                                                 rat_output_A=A_output_s, parallel=True, steps_size=step_size,
                                                 pol='s')

    assert approx(A_output == A_output_s)

    assert approx(result["absorption"] == result_s["absorption"])


@mark.skipif(sys.platform != "linux", reason="Only works under linux")
def test_rcwa_polygon():
    import numpy as np

    from solcore import material, si
    from solcore.structure import Layer, Structure
    from solcore.absorption_calculator.rigorous_coupled_wave import calculate_rat_rcwa, \
        calculate_absorption_profile_rcwa

    T = 300
    Air = material("Air")(T=T)
    TiO2 = material("TiO2", sopra=True)(T=T)  # for the nanoparticles
    GaAs = material("GaAs")(T=T)
    th = 50

    NP_layer_square = Layer(
        si(th, "nm"),
        Air,
        geometry=[{"type": "rectangle", "mat": TiO2, "center": (200, 200), "halfwidths": [50, 50], "angle": 0}],
    )
    np_struct_square = Structure([NP_layer_square])

    NP_layer_polygon = Layer(
        si(th, "nm"),
        Air,
        geometry=[{"type": "polygon", "mat": TiO2, "center": (200, 200),
                   "vertices": ((150, 150), (250,150), (250, 250), (150, 250)),
                                "angle": 0}],
    )
    np_struct_polygon = Structure([NP_layer_polygon])

    wl = np.linspace(300, 1000, 10)

    step_size = 2

    rat_np_sq = calculate_rat_rcwa(
        np_struct_square, size=((400, 0), (0, 400)), orders=10, wavelength=wl, substrate=GaAs, incidence=Air,
        pol=[1, 0]
    )

    A_output_sq = rat_np_sq['A_pol']
    result_square = calculate_absorption_profile_rcwa(np_struct_square, size=((400, 0), (0, 400)), orders=10, wavelength=wl,
                                                 rat_output_A=A_output_sq, parallel=True, steps_size=step_size,
                                                 pol=[1, 0])

    rat_np_pol = calculate_rat_rcwa(
        np_struct_polygon, size=((400, 0), (0, 400)), orders=10, wavelength=wl, substrate=GaAs, incidence=Air,
        pol=[1, 0]
    )

    A_output_pol = rat_np_pol['A_pol']
    result_polygon = calculate_absorption_profile_rcwa(np_struct_polygon, size=((400, 0), (0, 400)), orders=10, wavelength=wl,
                                                 rat_output_A=A_output_pol, parallel=True, steps_size=step_size,
                                                 pol=[1, 0])



    assert approx(A_output_sq == A_output_pol)

    assert approx(result_polygon["absorption"] == result_square["absorption"])


@mark.skipif(sys.platform != "linux", reason="Only works under linux")
@mark.parametrize("pol", ['s', 'p', 'u'])
@mark.parametrize("angle", [0, 60])
def test_tmm_rcwa_structure_comparison(pol, angle):
    import numpy as np
    from solcore import si, material
    from solcore.structure import Layer
    from solcore.solar_cell import SolarCell
    from solcore.optics.tmm import calculate_rat, OptiStack
    from solcore.optics.rcwa import calculate_rat_rcwa

    InGaP = material('GaInP')(In=0.5)
    GaAs = material('GaAs')()
    Ge = material('Ge')()
    Ag = material('Ag')()
    Air = material('Air')()

    Al2O3 = material('Al2O3')()

    wavelengths = np.linspace(250, 1900, 500)

    size = ((100, 0), (0, 100))

    ARC = [Layer(si('80nm'), Al2O3)]

    solar_cell = SolarCell(ARC + [Layer(material=InGaP, width=si('400nm')),
                                  Layer(material=GaAs, width=si('4000nm')),
                                  Layer(material=Ge, width=si('3000nm'))], substrate=Ag)

    solar_cell_OS = OptiStack(solar_cell, no_back_reflection=False, substrate=solar_cell.substrate)


    rcwa_result = calculate_rat_rcwa(solar_cell, size, 2, wavelengths, Air, Ag, theta=angle, phi=angle,
                                     pol=pol, parallel=True)
    tmm_result = calculate_rat(solar_cell_OS, wavelengths, angle, pol, no_back_reflection=False)

    assert tmm_result['A_per_layer'][1:-1] == approx(rcwa_result['A_per_layer'].T)
    assert tmm_result['R'] == approx(rcwa_result['R'])
    assert tmm_result['T'] == approx(rcwa_result['T'])

    assert np.sum(tmm_result['A_per_layer'][1:-1].T, 1) + tmm_result['R'] + tmm_result['T'] == approx(1)
    assert np.sum(rcwa_result['A_per_layer'], 1) + rcwa_result['R'] + rcwa_result['T'] == approx(1)


@mark.skipif(sys.platform != "linux", reason="Only works under linux")
@mark.parametrize("pol", ['s', 'p', 'u'])
@mark.parametrize("angle", [0, 60])
def test_tmm_rcwa_structure_profile_comparison(pol, angle):
    import numpy as np
    from solcore import si, material
    from solcore.structure import Layer
    from solcore.solar_cell import SolarCell
    from solcore.optics.tmm import calculate_rat, calculate_absorption_profile, OptiStack
    from solcore.optics.rcwa import calculate_rat_rcwa, calculate_absorption_profile_rcwa

    InGaP = material('GaInP')(In=0.5)
    GaAs = material('GaAs')()
    Ge = material('Ge')()
    Ag = material('Ag')()
    Air = material('Air')()

    wavelengths = np.linspace(250, 1900, 8)

    size = ((100, 0), (0, 100))

    solar_cell = SolarCell([Layer(material=InGaP, width=si('400.5nm')),
                                  Layer(material=GaAs, width=si('500nm')),
                                  Layer(material=Ge, width=si('1000nm'))], substrate=Ag)

    solar_cell_OS = OptiStack(solar_cell, no_back_reflection=False, substrate=solar_cell.substrate)

    rcwa_result = calculate_rat_rcwa(solar_cell, size, 2, wavelengths, Air, Ag, theta=angle, phi=angle,
                                     pol=pol, parallel=True)
    tmm_result = calculate_rat(solar_cell_OS, wavelengths, angle, pol, no_back_reflection=False)

    tmm_profile = calculate_absorption_profile(solar_cell_OS, wavelengths, no_back_reflection=False,
                                               angle=angle, pol=pol, RAT_out=tmm_result, steps_size=2)

    rcwa_profile = calculate_absorption_profile_rcwa(solar_cell, size, 2, wavelengths, rcwa_result['A_pol'],
                                                     theta=angle, phi=angle, pol=pol, incidence=Air,
                                                     substrate=solar_cell.substrate, parallel=True, steps_size=2)


    assert tmm_profile['position'] == approx(rcwa_profile['position'], abs=2e-6)
    assert tmm_profile['absorption'] == approx(rcwa_profile['absorption'], abs=2e-6)


import solcore.science_tracker as st

st.science_reference('InP refractive index', 'S. Adachi. Optical dispersion relations for GaP, GaAs, GaSb, InP, InAs, InSb, AlxGa1−xAs, and In1−xGaxAsyP1−y, J. Appl. Phys. 66, 6030-6040 (1989)')
st.science_reference('AlInP refractive index', 'E. Ochoa-Martínez, L. Barrutia, M. Ochoa, E. Barrigón, I. García, I. Rey-Stolle, C. Algora, P. Basa, G. Kronome, and M. Gabás, “Refractive indexes and extinction coefficients of n- and p-type doped GaInP, AlInP and AlGaInP for multijunction solar cells,” Solar Energy Materials and Solar Cells, vol. 174, pp. 388–396, Jan. 2018.')
st.science_reference('AlP refractive index. Cut to bandgap.', 'V. Emberger, F. Hatami, W. Ted Masselink, and S. Peters, “AlP/GaP distributed Bragg reflectors,” Appl. Phys. Lett., vol. 103, no. 3, pp. 031101–4, Jul. 2013.')


import numpy as np
from scipy.interpolate import interp1d
from scipy.integrate import solve_bvp

from solcore.constants import kb, q
from solcore.science_tracker import science_reference
from solcore.state import State
from solcore.light_source import LightSource


def identify_layers(junction):
    # First we have to figure out if we are talking about a PN, NP, PIN or NIP junction
    # We search for the emitter and check if it is n-type or p-type
    idx = 0
    pn_or_np = 'pn'
    homojunction = True

    for layer in junction:
        if layer.role.lower() != 'emitter':
            idx += 1
        else:
            Na = 0
            Nd = 0
            if hasattr(layer.material, 'Na'): Na = layer.material.Na
            if hasattr(layer.material, 'Nd'): Nd = layer.material.Nd
            if Na < Nd:
                pn_or_np = "np"
                nRegion = junction[idx]
            else:
                pRegion = junction[idx]

            id_top = idx

            break

    # Now we check for an intrinsic region and, if there is, for the base.
    if junction[idx + 1].role.lower() == 'intrinsic':
        iRegion = junction[idx + 1]

        if junction[idx + 2].role.lower() == 'base':
            if pn_or_np == "pn":
                nRegion = junction[idx + 2]

            else:
                pRegion = junction[idx + 2]

            id_bottom = idx + 2
            homojunction = homojunction and nRegion.material.material_string == pRegion.material.material_string
            homojunction = homojunction and nRegion.material.material_string == iRegion.material.material_string

        else:
            raise RuntimeError(
                'ERROR processing junctions: A layer following the "intrinsic" layer must be defined as '
                '"base".')

    # If there is no intrinsic region, we check directly the base
    elif junction[idx + 1].role.lower() == 'base':
        if pn_or_np == "pn":
            nRegion = junction[idx + 1]

        else:
            pRegion = junction[idx + 1]

        iRegion = None

        id_bottom = idx + 1
        homojunction = homojunction and nRegion.material.material_string == pRegion.material.material_string

    else:
        raise RuntimeError(
            'ERROR processing junctions: A layer following the "emitter" must be defined as "intrinsic"'
            'or "base".')

    # We assert that we are really working with an homojunction
    assert homojunction, 'ERROR: The DA solver only works with homojunctions, for now.'


    return id_top, id_bottom, pRegion, nRegion, iRegion, pn_or_np


def identify_parameters(junction, T, pRegion, nRegion, iRegion):

    kbT = kb * T

    xp = pRegion.width
    xn = nRegion.width
    xi = 0 if iRegion is None else iRegion.width

    sn = 0 if not hasattr(junction, "sn") else junction.sn
    sp = 0 if not hasattr(junction, "sp") else junction.sp

    # Now we have to get all the material parameters needed for the calculation
    if hasattr(junction, "permittivity"):
        es = junction.permittivity
    else:
        es = nRegion.material.permittivity  # equal for n and p.  I hope.

    # For the diffusion length, subscript n and p refer to the carriers, electrons and holes
    if hasattr(junction, "ln"):
        ln = junction.ln
    else:
        ln = pRegion.material.electron_diffusion_length

    if hasattr(junction, "lp"):
        lp = junction.lp
    else:
        lp = nRegion.material.hole_diffusion_length

    # For the diffusion coefficient, n and p refer to the regions, n side and p side. Yeah, it's confusing...
    if hasattr(junction, "mup"):
        dp = junction.mup * kbT / q
    else:
        dp = pRegion.material.electron_mobility * kbT / q

    if hasattr(junction, "mun"):
        dn = junction.mun * kbT / q
    else:
        dn = nRegion.material.hole_mobility * kbT / q

    ni = nRegion.material.ni

    Na = pRegion.material.Na
    Nd = nRegion.material.Nd

    return xn, xp, xi, sn, sp, ln, lp, dn, dp, Nd, Na, ni, es


def iv_depletion(junction, options):
    """ Calculates the IV curve of a junction object using the depletion approximation as described in J. Nelson, “The Physics of Solar Cells”, Imperial College Press (2003). The junction is then updated with an "iv" function that calculates the IV curve at any voltage.

    :param junction: A junction object.
    :param options: Solver options.
    :return: None.
    """

    science_reference('Depletion approximation',
                      'J. Nelson, “The Physics of Solar Cells”, Imperial College Press (2003).')

    junction.voltage = options.internal_voltages
    T = options.T
    kbT = kb * T

    id_top, id_bottom, pRegion, nRegion, iRegion, pn_or_np = identify_layers(junction)
    xn, xp, xi, sn, sp, ln, lp, dn, dp, Nd, Na, ni, es = identify_parameters(junction, T, pRegion, nRegion, iRegion)

    niSquared = ni**2

    Vbi = (kbT / q) * np.log(Nd * Na / niSquared) if not hasattr(junction, "Vbi") else junction.Vbi  # Jenny p146

    #Na, Nd, ni, niSquared, xi, ln, lp, xn, xp, sn, sp, dn, dp, es, id_top, id_bottom, Vbi, pn_or_np = process_junction(junction, options)

    R_shunt = min(junction.R_shunt, 1e14) if hasattr(junction, 'R_shunt') else 1e14

    # And now we account for the possible applied voltage, which can be, at most, equal to Vbi
    V = np.where(junction.voltage < Vbi - 0.001, junction.voltage, Vbi - 0.001)

    wn, wp = get_depletion_widths(junction, es, Vbi, V, Na, Nd, xi)

    w = wn + wp + xi

    # Now it is time to calculate currents
    if pn_or_np == "pn":
        l_top, l_bottom = ln, lp
        x_top, x_bottom = xp, xn
        w_top, w_bottom = wp, wn
        s_top, s_bottom = sp, sn
        d_top, d_bottom = dp, dn
        min_top, min_bot = niSquared / Na, niSquared / Nd
    else:
        l_bottom, l_top = ln, lp
        x_bottom, x_top = xp, xn
        w_bottom, w_top = wp, wn
        s_bottom, s_top = sp, sn
        d_bottom, d_top = dp, dn
        min_bot, min_top = niSquared / Na, niSquared / Nd

    JtopDark = get_j_dark(x_top, w_top, l_top, s_top, d_top, V, min_top, T)
    JbotDark = get_j_dark(x_bottom, w_bottom, l_bottom, s_bottom, d_bottom, V, min_bot, T)

    # hereby we define the subscripts to refer to the layer in which the current is generated:
    if pn_or_np == "pn":
        JnDark, JpDark = JbotDark, JtopDark
    else:
        JpDark, JnDark = JbotDark, JtopDark

    # These might not be the right lifetimes. Actually, they are not as they include all recombination processes, not
    # just SRH recombination, which is what the equation in Jenny, p159 refers to. Let´ leave them, for now.
    lifetime_n = ln ** 2 / dn
    lifetime_p = lp ** 2 / dp  # Jenny p163

    # Here we use the full version of the SRH recombination term as calculated by Sah et al. Works for positive bias
    # and moderately negative ones.

    science_reference('SRH current term.',
                      'C. T. Sah, R. N. Noyce, and W. Shockley, “Carrier Generation and Recombination in P-N Junctions and P-N Junction Characteristics,” presented at the Proceedings of the IRE, 1957, vol. 45, no. 9, pp. 1228–1243.')
    Jrec = get_Jsrh(ni, V, Vbi, lifetime_p, lifetime_n, w, kbT)

    J_sc_top = 0
    J_sc_bot = 0
    J_sc_scr = 0

    if options.light_iv:

        widths = []
        for layer in junction:
            widths.append(layer.width)

        cum_widths = np.cumsum([0] + widths)

        g = junction.absorbed
        wl = options.wavelength
        wl_sp, ph = options.light_source.spectrum(output_units='photon_flux_per_m', x=wl)
        id_v0 = np.argmin(abs(V))

        # The contribution from the Emitter (top side).
        xa = cum_widths[id_top]
        xb = cum_widths[id_top + 1] - w_top[id_v0]

        deriv = get_J_sc_diffusion(xa, xb, g, d_top, l_top, min_top, s_top, wl, ph, side='top')
        J_sc_top = q * d_top * abs(deriv)

        # The contribution from the Base (bottom side).
        xa = cum_widths[id_bottom] + w_bottom[id_v0]
        xb = cum_widths[id_bottom + 1]
        deriv = get_J_sc_diffusion(xa, xb, g, d_bottom, l_bottom, min_bot, s_bottom, wl, ph, side='bottom')
        J_sc_bot = q * d_bottom * abs(deriv)

        # The contribution from the SCR (includes the intrinsic region, if present).
        xa = cum_widths[id_top + 1] - w_top[id_v0]
        xb = cum_widths[id_bottom] + w_bottom[id_v0]
        J_sc_scr = q * get_J_sc_SCR(xa, xb, g, wl, ph)

    # And, finally, we output the currents
    junction.current = Jrec + JnDark + JpDark + V / R_shunt - J_sc_top - J_sc_bot - J_sc_scr
    junction.iv = interp1d(junction.voltage, junction.current, kind='linear', bounds_error=False, assume_sorted=True,
                           fill_value=(junction.current[0], junction.current[-1]))
    junction.region_currents = State({"Jn_dif": JnDark, "Jp_dif": JpDark, "Jscr_srh": Jrec,
                                      "J_sc_top": J_sc_top, "J_sc_bot": J_sc_bot, "J_sc_scr": J_sc_scr})


def get_j_dark(x, w, l, s, d, V, minority, T):
    """
    :param x: width of top junction
    :param w: depletion width in top junction
    :param l: diffusion length
    :param s: surface recombination velocity
    :param d: diffusion coefficient
    :param V: voltage
    :param minority: minority carrier density
    :param T: Temperature

    :return: J_top_dark
    """
    # We calculate some fractions

    harg = (x - w) / l
    sinh_harg = np.sinh(harg)
    cosh_harg = np.cosh(harg)
    lsod = (l * s) / d

    # And then we are ready to calculate the different currents
    # Missing the voltage dependent part of these equations.
    # They should be 6.34 and 6.39, not 6.62 and 6.63

    J_dark = (q * d * minority / l) * (np.exp(q * V / kb / T) - 1) * \
                 ((lsod * cosh_harg + sinh_harg) / (lsod * sinh_harg + cosh_harg))

    return J_dark


def get_Jsrh(ni, V, Vbi, tp, tn, w, kbT, dEt=0):
    science_reference('SRH current term.',
                      'C. T. Sah, R. N. Noyce, and W. Shockley, “Carrier Generation and Recombination in P-N Junctions and P-N Junction Characteristics,” presented at the Proceedings of the IRE, 1957, vol. 45, no. 9, pp. 1228–1243.')

    out = np.zeros(V.shape)

    m = V >= -1120 * kbT / q
    out[m] = forward(ni, V[m], Vbi, tp, tn, w[m], kbT)

    return out


def forward(ni, V, Vbi, tp, tn, w, kbT, dEt=0):
    """ Equation 27 of Sah's paper. Strictly speaking, it is not valid for intermediate negative bias. """

    J0 = 2 * q * ni * w / np.sqrt(tn * tp)
    f = factor(V, Vbi, tp, tn, kbT, dEt)
    out = J0 * np.sinh(q * V / (2 * kbT)) / (q * (Vbi - V) / kbT) * f

    return out


def factor(V, Vbi, tp, tn, kbT, dEt=0):
    """ The integral of Eq. 27 in Sah's paper. While it is coninuum (in principle) it has to be done in two parts.
    (or three) """
    trap = q * dEt / kbT + np.log(tp / tn) / 2
    b = np.exp(-q * V / kbT / 2) * np.cosh(trap)

    z1 = np.sqrt(tp / tn) * np.exp(-q * (Vbi - V) / kbT / 2)
    z2 = np.sqrt(tp / tn) * np.exp(q * (Vbi - V) / kbT / 2)

    out = np.zeros(b.shape)

    # For b values < 1
    m = b < 1
    top = np.arctan((b[m] + z2[m]) / np.sqrt(1 - b[m] ** 2))
    bot = np.arctan((b[m] + z1[m]) / np.sqrt(1 - b[m] ** 2))

    out[m] = (top - bot) / np.sqrt(1 - b[m] ** 2)

    # For b values >= 1 and <=1e7
    m = (b >= 1) * (b <= 1e7)
    xx = b[m]
    yy = np.sqrt(xx ** 2 - 1)

    top = np.log(abs(z2[m] - yy + xx) / abs(z2[m] + yy + xx))
    bot = np.log(abs(z1[m] - yy + xx) / abs(z1[m] + yy + xx))

    out[m] = (top - bot) / (2 * yy)

    # For b values > 1e7
    m = b >= 1e7

    top = np.log(z2[m]) - np.log(z2[m] + 2 * b[m])
    bot = np.log(z1[m]) - np.log(z1[m] + 2 * b[m])

    out[m] = (top - bot) / (2 * b[m])

    return out


def get_J_sc_diffusion(xa, xb, g, D, L, y0, S, wl, ph, side='top'):
    """
    :param xa:
    :param xb:
    :param g:
    :param D:
    :param L:
    :param y0:
    :param S:
    :param wl:
    :param ph:
    :param side:

    :return: out
    """

    zz = np.linspace(xa, xb, 1001, endpoint=False)
    gg = g(zz) * ph

    g_vs_z = np.trapz(gg, wl, axis=1)

    g_vs_z[np.isnan(g_vs_z)] = 0

    A = lambda x: np.interp(x, zz, g_vs_z) / D + y0 / L ** 2

    def fun(x, y):
        out1 = y[1]
        out2 = y[0] / L ** 2 - A(x)
        return np.vstack((out1, out2))

    if side == 'top':
        def bc(ya, yb):
            left = ya[1] - S / D * (ya[0] - y0)
            right = yb[0]
            return np.array([left, right])
    else:
        def bc(ya, yb):
            left = ya[0]
            right = yb[1] - S / D * (yb[0] - y0)
            return np.array([left, right])

    guess = y0 * np.ones((2, zz.size))
    guess[1] = np.zeros_like(guess[0])

    solution = solve_bvp(fun, bc, zz, guess)

    if side == 'top':
        out = solution.y[1][-1]
    else:
        out = solution.y[1][0]

    return out


def get_J_sc_SCR(xa, xb, g, wl, ph):
    zz = np.linspace(xa, xb, 1001, endpoint=False)
    gg = g(zz) * ph
    out = np.trapz(np.trapz(gg, wl, axis=1), zz)

    return out


def qe_depletion(junction, options):
    """ Calculates the QE curve of a junction object using the depletion approximation as described in J. Nelson, “The Physics of Solar Cells”, Imperial College Press (2003). The junction is then updated with an "iqe" and several "eqe" functions that calculates the QE curve at any wavelength.

    :param junction: A junction object.
    :param options: Solver options.
    :return: None.
    """

    science_reference('Depletion approximation',
                      'J. Nelson, “The Physics of Solar Cells”, Imperial College Press (2003).')


    # First we have to figure out if we are talking about a PN, NP, PIN or NIP junction
    T = options.T
    kbT = kb * T

    id_top, id_bottom, pRegion, nRegion, iRegion, pn_or_np = identify_layers(junction)
    xn, xp, xi, sn, sp, ln, lp, dn, dp, Nd, Na, ni, es = identify_parameters(junction, T, pRegion, nRegion, iRegion)

    niSquared = ni ** 2

    Vbi = (kbT / q) * np.log(Nd * Na / niSquared) if not hasattr(junction, "Vbi") else junction.Vbi  # Jenny p146

    wn, wp = get_depletion_widths(junction, es, Vbi, 0, Na, Nd, xi)

    # Now it is time to calculate currents
    if pn_or_np == "pn":
        l_top, l_bottom = ln, lp
        w_top, w_bottom = wp, wn
        s_top, s_bottom = sp, sn
        d_top, d_bottom = dp, dn
        min_top, min_bot = niSquared / Na, niSquared / Nd
    else:
        l_bottom, l_top = ln, lp
        w_bottom, w_top = wp, wn
        s_bottom, s_top = sp, sn
        d_bottom, d_top = dp, dn
        min_bot, min_top = niSquared / Na, niSquared / Nd

    widths = []
    for layer in junction:
        widths.append(layer.width)

    cum_widths = np.cumsum([0] + widths)

    g = junction.absorbed
    wl = options.wavelength
    wl_sp, ph = LightSource(source_type='black body', x=wl, T=6000).spectrum(output_units='photon_flux_per_m', x=wl)

    # The contribution from the Emitter (top side).
    xa = cum_widths[id_top]
    xb = cum_widths[id_top + 1] - w_top

    deriv = get_J_sc_diffusion_vs_WL(xa, xb, g, d_top, l_top, min_top, s_top, wl, ph, side='top')
    j_sc_top = d_top * abs(deriv)


    # The contribution from the Base (bottom side).
    xa = cum_widths[id_bottom] + w_bottom
    xb = cum_widths[id_bottom + 1]

    deriv = get_J_sc_diffusion_vs_WL(xa, xb, g, d_bottom, l_bottom, min_bot, s_bottom, wl, ph, side='bottom')
    j_sc_bot = d_bottom * abs(deriv)

    # The contribution from the SCR (includes the intrinsic region, if present).
    xa = cum_widths[id_top + 1] - w_top
    xb = cum_widths[id_bottom] + w_bottom
    j_sc_scr = get_J_sc_SCR_vs_WL(xa, xb, g, wl, ph)

    # The total light absorbed, but not necessarily collected, is:
    xa = cum_widths[id_top]
    xb = cum_widths[id_bottom + 1]
    zz = np.linspace(xa, xb, 10001)
    gg = g(zz) * ph
    current_absorbed = np.trapz(gg, zz, axis=0)

    # Now, we put everything together
    j_sc = j_sc_top + j_sc_bot + j_sc_scr

    eqe = j_sc / ph
    eqe_emitter = j_sc_top / ph
    eqe_base = j_sc_bot / ph
    eqe_scr = j_sc_scr / ph

    iqe =  j_sc / current_absorbed
    iqe[np.isnan(iqe)] = 0 # if zero current_absorbed, get NaN in previous line; want 0 IQE

    junction.iqe = interp1d(wl, iqe)

    junction.eqe = interp1d(wl, eqe, kind='linear', bounds_error=False, assume_sorted=True,
                            fill_value=(eqe[0], eqe[-1]))
    junction.eqe_emitter = interp1d(wl, eqe_emitter, kind='linear', bounds_error=False, assume_sorted=True,
                                    fill_value=(eqe_emitter[0], eqe_emitter[-1]))
    junction.eqe_base = interp1d(wl, eqe_base, kind='linear', bounds_error=False, assume_sorted=True,
                                 fill_value=(eqe_base[0], eqe_base[-1]))
    junction.eqe_scr = interp1d(wl, eqe_scr, kind='linear', bounds_error=False, assume_sorted=True,
                                fill_value=(eqe_scr[0], eqe_scr[-1]))

    junction.qe = State({'WL': wl, 'IQE': junction.iqe(wl), 'EQE': junction.eqe(wl), 'EQE_emitter': junction.eqe_emitter(wl),
                         'EQE_base': junction.eqe_base(wl), 'EQE_scr': junction.eqe_scr(wl)})

def get_J_sc_SCR_vs_WL(xa, xb, g, wl, ph):
    zz = np.linspace(xa, xb, 1001, endpoint=False)
    gg = g(zz) * ph
    out = np.trapz(gg, zz, axis=0)

    return out


def get_J_sc_diffusion_vs_WL(xa, xb, g, D, L, y0, S, wl, ph, side='top'):
    zz = np.linspace(xa, xb, 1001, endpoint=False) # excluding the last point - depending on the mesh/floating point errors, sometimes this is actually in the next layer
    gg = g(zz) * ph
    out = np.zeros_like(wl)

    for i in range(len(wl)):

        if np.all(gg[:,i] == 0): # no reason to solve anything if no generation at this wavelength
            out[i] = 0

        else:
            A = lambda x: np.interp(x, zz, gg[:, i]) / D + y0 / L ** 2

            def fun(x, y):
                out1 = y[1]
                out2 = y[0] / L ** 2 - A(x)
                return np.vstack((out1, out2))

            if side == 'top':
                def bc(ya, yb):
                    left = ya[1] - S / D * (ya[0] - y0)
                    right = yb[0]
                    return np.array([left, right])
            else:
                def bc(ya, yb):
                    left = ya[0]
                    right = yb[1] - S / D * (yb[0] - y0)
                    return np.array([left, right])

            guess = y0 * np.ones((2, zz.size))
            guess[1] = np.zeros_like(guess[0])
            solution = solve_bvp(fun, bc, zz, guess)

            if side == 'top':
                out[i] = solution.y[1][-1]
            else:
                out[i] = solution.y[1][0]

    return out


def get_depletion_widths(junction, es, Vbi, V, Na, Nd, xi):

    if not hasattr(junction, "wp") or not hasattr(junction, "wn"):

        if hasattr(junction, "depletion_approximation") and junction.depletion_approximation == "one-sided abrupt":
            print("using one-sided abrupt junction approximation for depletion width")
            one_sided = True
        else:
            one_sided = False


        if one_sided:
            science_reference("Sze abrupt junction approximation",
                              "Sze: The Physics of Semiconductor Devices, 2nd edition, John Wiley & Sons, Inc (2007)")
            wn = np.sqrt(2 * es * (Vbi - V) / (q * Nd))
            wp = np.sqrt(2 * es * (Vbi - V) / (q * Na))

        else:
            wn = (-xi + np.sqrt(xi ** 2 + 2. * es * (Vbi - V) / q * (1 / Na + 1 / Nd))) / (1 + Nd / Na)
            wp = (-xi + np.sqrt(xi ** 2 + 2. * es * (Vbi - V) / q * (1 / Na + 1 / Nd))) / (1 + Na / Nd)

    wn = wn if not hasattr(junction, "wn") else junction.wn
    wp = wp if not hasattr(junction, "wp") else junction.wp

    return wn, wp


import numpy as np
import sys
from scipy.sparse import dia_matrix
from scipy.sparse.linalg import eigs
from solcore.constants import q, hbar, electron_mass as m0

from operator import itemgetter

sort_simultaneous = lambda *lists: [list(x) for x in zip(*sorted(zip(*lists), key=itemgetter(0)))]


def alfa(Eband, m):
    return (1 - m) ** 2 * q / Eband


def kp4x4(mat, host_material):
    """ Calculates the bandedges and bulk effective masses of a material under biaxial strain at the gamma point.
    SO band is ignored.

    This function provides the position at the zone center of the C, HH and LH bands of a material under biaxial strain.
    Both, the paralell and transverse effective masses (with respect the Z axis) are provided. SO band is ignored,
    therefore the masses are independent of strain. """

    Delta = mat.spin_orbit_splitting
    Eg = mat.band_gap
    Ev0 = mat.valence_band_offset
    Ec0 = mat.valence_band_offset + Eg
    a0 = mat.lattice_constant
    g1 = mat.gamma1
    g2 = mat.gamma2
    g3 = mat.gamma3
    ac = mat.a_c
    av = abs(mat.a_v)
    b = mat.b
    me = mat.eff_mass_electron_Gamma

    # Strain parameters
    exx = (host_material.lattice_constant - a0) / a0
    ezz = - 2 * mat.c12 / mat.c11 * exx

    Oe = ac * (exx + exx + ezz)
    Pe = -av * (exx + exx + ezz)
    Qe = -b / 2 * (exx + exx - 2 * ezz)

    Ec = Ec0 + Oe
    Evhh = Ev0 - Pe - Qe
    Evlh = Ev0 - Pe + Qe

    # Effective masses paralell to the Z axis
    mhh_p = 1. / (g1 - 2 * g2)
    mlh_p = 1. / (g1 + 2 * g2)

    # Effective masses transverse to the Z axis, that is, in the plane of the strain
    mhh_t = 1. / (g1 + g2)
    mlh_t = 1. / (g1 - g2)

    return Ec, Evhh, Evlh, me * m0, mhh_p * m0, mlh_p * m0, mhh_t * m0, mlh_t * m0


def kp6x6(mat, host_material):
    """ Calculates the bandedges and bulk effective masses of a material under biaxial strain at the gamma point
    including the SO band.

    This function provides the position at the zone center of the C, HH, LH and SO bands of a material under biaxial
    strain. Both, the paralell and transverse effective masses (with respect the Z axis) are provided."""

    Delta = mat.spin_orbit_splitting
    Eg = mat.band_gap
    Ev0 = mat.valence_band_offset
    Ec0 = mat.valence_band_offset + Eg
    a0 = mat.lattice_constant
    g1 = mat.gamma1
    g2 = mat.gamma2
    g3 = mat.gamma3
    ac = mat.a_c
    av = abs(mat.a_v)
    b = mat.b
    me = mat.eff_mass_electron_Gamma

    # Strain parameters
    exx = (host_material.lattice_constant - a0) / a0
    ezz = - 2 * mat.c12 / mat.c11 * exx

    Oe = ac * (exx + exx + ezz)
    Pe = -av * (exx + exx + ezz)
    Qe = -b / 2 * (exx + exx - 2 * ezz)

    x = Qe / Delta
    if x == 0: x = 1e-10
    com = lambda s: x - 1 + s * np.sqrt(1 + 2 * x + 9 * x ** 2)
    f = lambda s: (2 * x * (1 + 1.5 * com(s)) + 6 * x ** 2) / (
        0.75 * com(s) ** 2 + com(s) - 3 * x ** 2)  # s=+1 for LH and s=-1 for SO

    Ec = Ec0 + Oe
    Evhh = Ev0 - Pe - Qe
    Evlh = Ev0 - Pe + 0.5 * (Qe - Delta + np.sqrt(Delta ** 2 + 2 * Delta * Qe + 9 * Qe ** 2))
    Evso = Ev0 - Pe + 0.5 * (Qe - Delta - np.sqrt(Delta ** 2 + 2 * Delta * Qe + 9 * Qe ** 2))

    # Effective masses paralell to the Z axis
    mhh_p = 1. / (g1 - 2 * g2)
    mlh_p = 1. / (g1 + 2 * f(+1) * g2)
    mso_p = 1. / (g1 + 2 * f(-1) * g2)

    # Effective masses transverse to the Z axis, that is, in the plane of the strain
    mhh_t = 1. / (g1 + g2)
    mlh_t = 1. / (g1 - f(+1) * g2)
    mso_t = 1. / (g1 - f(-1) * g2)

    return Ec, Evhh, Evlh, Evso, me * m0, mhh_p * m0, mlh_p * m0, mso_p * m0, mhh_t * m0, mlh_t * m0, mso_t * m0


def fill_hamiltonian_holes_4x4(A1, A2, B1, B2, C, D, delta, block):
    # Number of meshpoints in z direction. The number of equations to solve is 2N
    N = len(A1)

    # Values of the diagonals, depending on the equation: 1, 2, 3 or 4
    ll14 = lambda i: (A2[i + 1] - 4 * A2[i] - A2[i - 1]) / (4 * delta ** 2)
    l13 = lambda i: D[i] / (2 * delta)
    d14 = lambda i: 2 * A2[i] / delta ** 2 + A1[i]
    u13 = lambda i: C[i] - (D[i + 1] - D[i - 1]) / (4 * delta)
    uu14 = lambda i: - (A2[i + 1] + 4 * A2[i] - A2[i - 1]) / (4 * delta ** 2)
    uuu13 = lambda i: - D[i] / (2 * delta)

    lll24 = lambda i: - D[i] / (2 * delta)
    ll23 = lambda i: (B2[i + 1] - 4 * B2[i] - B2[i - 1]) / (4 * delta ** 2)
    l24 = lambda i: C[i] + (D[i + 1] - D[i - 1]) / (4 * delta)
    d23 = lambda i: 2 * B2[i] / delta ** 2 + B1[i]
    u24 = lambda i: D[i] / (2 * delta)
    uu23 = lambda i: - (B2[i + 1] + 4 * B2[i] - B2[i - 1]) / (4 * delta ** 2)

    lll = np.zeros(2 * N)
    ll = np.zeros(2 * N)
    l = np.zeros(2 * N)
    d = np.zeros(2 * N)
    u = np.zeros(2 * N)
    uu = np.zeros(2 * N)
    uuu = np.zeros(2 * N)

    if block is "U":
        # We write the equations for the UPPER 2x2 block of the Hamiltonian, for g1 and g2

        # We fill the first two equations separatelly
        d[0] = d14(0)
        u[0] = C[0] - (D[1] - D[0]) / (4 * delta)
        uu[0] = - (A2[1] + 3 * A2[0]) / (4 * delta ** 2)
        uuu[0] = uuu13(0)

        l[1] = C[0] + (D[1] - D[0]) / (4 * delta)
        d[1] = d23(0)
        u[1] = u24(0)
        uu[1] = - (B2[1] + 3 * B2[0]) / (4 * delta ** 2)

        for i in range(1, N - 1):
            # The first equation, for g1
            k = 2 * i

            ll[k] = ll14(i)
            l[k] = l13(i)
            d[k] = d14(i)
            u[k] = u13(i)
            uu[k] = uu14(i)
            uuu[k] = uuu13(i)

            # The second equation, for g2
            k = k + 1

            lll[k] = lll24(i)
            ll[k] = ll23(i)
            l[k] = l24(i)
            d[k] = d23(i)
            u[k] = u24(i)
            uu[k] = uu23(i)

        # We fill the last two equations separatelly
        ll[2 * (N - 1)] = (- 3 * A2[N - 1] - A2[N - 2]) / (4 * delta ** 2)
        l[2 * (N - 1)] = l13(N - 1)
        d[2 * (N - 1)] = d14(N - 1)
        u[2 * (N - 1)] = C[N - 1] - (D[N - 1] - D[N - 2]) / (4 * delta)

        lll[2 * (N - 1) + 1] = lll24(N - 1)
        ll[2 * (N - 1) + 1] = (- 3 * B2[N - 1] - B2[N - 2]) / (4 * delta ** 2)
        l[2 * (N - 1) + 1] = C[N - 1] + (D[N - 1] - D[N - 2]) / (4 * delta)
        d[2 * (N - 1) + 1] = d23(N - 1)

    else:
        # We write the equations for the LOWER 2x2 block of the Hamiltonian, for g3 and g4

        # We fill the first two equations separatelly
        d[0] = d23(0)
        u[0] = C[0] - (D[1] - D[0]) / (4 * delta)
        uu[0] = - (B2[1] + 3 * B2[0]) / (4 * delta ** 2)
        uuu[0] = uuu13(0)

        l[1] = C[0] + (D[1] - D[0]) / (4 * delta)
        d[1] = d14(0)
        u[1] = u24(0)
        uu[1] = - (A2[1] + 3 * A2[0]) / (4 * delta ** 2)

        for i in range(1, N - 1):
            # The first equation, for g3
            k = 2 * i

            ll[k] = ll23(i)
            l[k] = l13(i)
            d[k] = d23(i)
            u[k] = u13(i)
            uu[k] = uu23(i)
            uuu[k] = uuu13(i)

            # The second equation, for g4
            k = k + 1

            lll[k] = lll24(i)
            ll[k] = ll14(i)
            l[k] = l24(i)
            d[k] = d14(i)
            u[k] = u24(i)
            uu[k] = uu14(i)

        # We fill the last two equations separatelly

        ll[2 * (N - 1)] = (- 3 * B2[N - 1] - B2[N - 2]) / (4 * delta ** 2)
        l[2 * (N - 1)] = l13(N - 1)
        d[2 * (N - 1)] = d23(N - 1)
        u[2 * (N - 1)] = C[N - 1] - (D[N - 1] - D[N - 2]) / (4 * delta)

        lll[2 * (N - 1) + 1] = lll24(N - 1)
        ll[2 * (N - 1) + 1] = (- 3 * A2[N - 1] - A2[N - 2]) / (4 * delta ** 2)
        l[2 * (N - 1) + 1] = C[N - 1] + (D[N - 1] - D[N - 2]) / (4 * delta)
        d[2 * (N - 1) + 1] = d14(N - 1)

    lll = np.roll(lll, -3)
    ll = np.roll(ll, -2)
    l = np.roll(l, -1)
    u = np.roll(u, 1)
    uu = np.roll(uu, 2)
    uuu = np.roll(uuu, 3)

    offsets = [-3, -2, -1, 0, 1, 2, 3]
    diagonals = np.array([lll, ll, l, d, u, uu, uuu])

    output = dia_matrix((diagonals, offsets), shape=(2 * N, 2 * N))

    # View hamiltonian, good for debugging
    # import pylab
    # pylab.imshow(np.real(output.todense()))
    # pylab.show()

    return output


def solve_holes_QW_at_kt_4x4(kt, z, fhh, flh, g1, g2, g3, num=(10, 10), quasiconfined=0.0, symmetric=False):
    # Normalization factors
    N = len(z)
    L = max(z)
    delta = (z[1] - z[0]) / L
    kt = kt * L
    E0 = hbar ** 2 / (m0 * L ** 2)

    # We invert and shift the potential so that it is possitive with the minimum value around zero
    offset = min(np.amin(fhh), np.amin(flh))
    fhh = (offset - fhh) / E0
    flh = (offset - flh) / E0

    # Fill the components of the equations based on the effective masses, potential and k value
    A1 = 0.5 * (g1 + g2) * kt ** 2 + fhh
    B1 = 0.5 * (g1 - g2) * kt ** 2 + flh
    A2 = 0.5 * (g1 - 2 * g2)
    B2 = 0.5 * (g1 + 2 * g2)
    C = np.sqrt(3) / 2 * 0.5 * (g2 + g3) * kt ** 2
    D = np.sqrt(3) * g3 * kt

    # Using the shift-invert mode will help to speed up the algorithm by calculating the eigenvalues only inside the QW or very close
    # See explanation: https://docs.scipy.org/doc/scipy-0.15.1/reference/tutorial/arpack.html
    sigma = min(np.amin(fhh), np.amin(
        flh))  # The bottom of the potential (valence band edge) is the target energy for the eigenvalues

    # Allow for quasiconfined levels to go through. They can be discarded later with the filter, if necessary
    potmax = max(np.amax(fhh), np.amax(flh)) + quasiconfined * q / E0

    # Solutions for upper hamiltonian
    H_U = fill_hamiltonian_holes_4x4(A1, A2, B1, B2, C, D, delta, "U")
    E_U, Psi_U = eigs(H_U, k=num[0], which='LR', sigma=sigma)

    Psi_g1 = Psi_U[range(0, len(Psi_U), 2)]
    Psi_g2 = Psi_U[range(1, len(Psi_U), 2)]

    if quasiconfined < 0.0:
        potmax = max(E_U)

    confined_levels = [i for i, e in enumerate(E_U) if (e < potmax and i < num[1])]
    E_U = E_U[confined_levels].real
    idx_sorted = np.argsort(E_U)
    E_U = E_U[idx_sorted]
    Psi_g1 = np.array(Psi_g1[:, confined_levels]).transpose().real[idx_sorted]
    Psi_g2 = np.array(Psi_g2[:, confined_levels]).transpose().real[idx_sorted]

    norm1 = [(np.sqrt(np.trapz(p * p, x=z))) for p in Psi_g1]
    norm2 = [(np.sqrt(np.trapz(p * p, x=z))) for p in Psi_g2]
    # Normalise the wavefunction except if the norm is very small. That's the case for some E at k near 0 only
    for i in range(len(Psi_g1)):
        if norm1[i] > 0.01 * max(norm1):
            Psi_g1[i] = Psi_g1[i] / norm1[i]
        else:
            Psi_g1[i] = Psi_g1[i] * 0.0

        if norm2[i] > 0.01 * max(norm1):
            Psi_g2[i] = Psi_g2[i] / norm2[i]
        else:
            Psi_g2[i] = Psi_g2[i] * 0.0

    E_U = offset - E_U * E0

    # If the structure is not symmetric, we have to solve the lower hamiltonian, too
    if not symmetric:
        H_L = fill_hamiltonian_holes_4x4(A1, A2, B1, B2, C, D, delta, "L")
        E_L, Psi_L = eigs(H_L, k=num[0], which='LR', sigma=sigma)

        Psi_g3 = Psi_L[range(0, len(Psi_L), 2)]
        Psi_g4 = Psi_L[range(1, len(Psi_L), 2)]

        if quasiconfined < 0.0:
            potmax = max(E_L)

        confined_levels = [i for i, e in enumerate(E_L) if (e < potmax and i < num[1])]
        E_L = E_L[confined_levels].real
        Psi_g3 = np.array(Psi_g3[:, confined_levels]).transpose().real
        Psi_g4 = np.array(Psi_g4[:, confined_levels]).transpose().real

        norm3 = [(np.sqrt(np.trapz(p * p, x=z))) for p in Psi_g3]
        norm4 = [(np.sqrt(np.trapz(p * p, x=z))) for p in Psi_g4]
        for i in range(len(Psi_g3)):
            if norm3[i] > 0.01 * max(norm3):
                Psi_g3[i] = Psi_g3[i] / norm3[i]
            else:
                Psi_g3[i] = Psi_g3[i] * 0.0

            if norm4[i] > 0.01 * max(norm4):
                Psi_g4[i] = Psi_g4[i] / norm4[i]
            else:
                Psi_g4[i] = Psi_g4[i] * 0.0

        E_L = offset - E_L * E0

        return {"E_U": E_U,
                "E_L": E_L,
                "Psi_g1": Psi_g1,
                "Psi_g2": Psi_g2,
                "Psi_g3": Psi_g3,
                "Psi_g4": Psi_g4}

    else:
        # If the structure is symmetric, we simply duplicate the energies and reverse the wavefunctions
        return {"E_U": E_U,
                "Psi_g1": Psi_g1,
                "Psi_g2": Psi_g2}


def fill_hamiltonian_holes_6x6(A1, A2, B1, B2, C, D, delta, block):
    print('Fully coupled 6x6 kp solver not implemented yet. Only the band edges can be calculated. ')
    sys.exit(-1)


def solve_holes_QW_at_kt_6x6(kt, z, fhh, flh, g1, g2, g3, num=(10, 10), quasiconfined=0.0, symmetric=False):
    print('Fully coupled 6x6 kp solver not implemented yet. Only the band edges can be calculated. ')
    sys.exit(-1)


def solve_electrons_QW_at_kt_parabolic(kt, z, fe, me, num=(10, 10), quasiconfined=0.0):
    """ Returns eignvalue and eigenvectors of an arbitrary potential.

    A tridiagonal matrix is constructed by writing the variable effective
    mass Schrodinger equation over a series of mesh points. The eigenvalues
    of the matrix correspond to the allowed energy levels of the system.

    The previous solver, eig, has been replaced by the spare matrix version, eigs, that is faster to compute


    "Varible effective mass Schordinger equation and tridiagonal solution method.",
    "Frensley, W. R. (1991). \
    Numerical evaluation of resonant states. \
    Superlattices and Microstructures, 11(3), 347350. \
    doi:10.1016/0749-6036(92)90396-M")

    """

    # Normalization factors
    L = max(z)
    dz = np.gradient(z / L)
    kt = kt * L
    E0 = hbar ** 2 / (m0 * L ** 2)

    # We shift the potential so that it is possitive with the minimum value around zero
    offset = np.amin(fe)
    fe = (fe - offset) / E0

    N = len(fe)
    m = me / m0

    # Vectorise effective mass differences to avoid a loop
    m = np.insert(m, (0, len(m)), (m[0], m[-1]))
    m_a = m[0:-2]  # m_(j-1) from Frensley, W. R. (1991)
    m_b = m[1:-1]  # m_j
    m_c = m[2:]  # m_(j+1)

    # These are the interior diagonals of equation 18 in the above ref.
    axis = 0.25 / dz ** 2 * (
    1 / m_a + 2 / m_b + 1 / m_c) + fe + 0.5 * kt ** 2 / m_b  # d_j from Frensley, W. R. (1991), modified to include kt and normalised variables
    upper = 0.25 / dz ** 2 * (1 / m_a + 1 / m_b)  # s_(j+1)
    lower = 0.25 / dz ** 2 * (1 / m_b + 1 / m_c)  # s_j

    index = (-1, 0, 1)
    diagonals = (-lower, axis, -upper)

    H = dia_matrix((diagonals, index), shape=(N, N))

    sigma = 0.0  # The bottom of the potential (conduction band) is the target energy for the eigenvalues

    # The heavy numerical calculation
    E, Psi = eigs(H, k=num[0], which='LR', sigma=sigma)

    # Allow for quasi confined levels to go through.
    potmax = max(fe) + quasiconfined * q / E0
    if quasiconfined < 0.0:
        potmax = max(E)

    confined_levels = [i for i, e in enumerate(E) if (e < potmax and i < num[1])]
    E = E[confined_levels].real
    Psi = np.array(Psi[:, confined_levels]).transpose()
    Psi = [(p / np.sqrt(np.trapz(p * p.conjugate(), x=z))) for p in Psi]

    E = E * E0 + offset

    # The solutions are degenerate: same energy
    return {"E_e": E,
            "Psi_e": Psi,
            "z": z}


def band_sorting(bands, symmetric=False):
    """ Sort the bands in C, HH and LH according to their character at k=0
    :param bands: A dictionary containing the bandstructure as calculated by solve_electrons_QW_at_kt_parabolic and
            solve_holes_QW_at_kt_4x4
    :return: A dictionary with the same input information but arranged in a different way, including labels
    """

    print("Organising the bands into E, HH and LH")

    Ee = []
    Ehh = []
    Elh = []
    psi_e = []
    psi_hh = []
    psi_lh = []

    # We vectorise the energies and the wavefunctions
    num_e = len(bands[0][1]["E_e"])
    num_hu = len(bands[0][1]["E_U"])
    kt = np.zeros(len(bands))
    electrons = np.zeros((len(bands), num_e))
    holes_U = np.zeros((len(bands), num_hu))
    Psi_e = np.zeros((len(bands), num_e, len(bands[0][1]["z"])))
    Psi_g1 = np.zeros((len(bands), num_hu, len(bands[0][1]["z"])))
    Psi_g2 = np.zeros((len(bands), num_hu, len(bands[0][1]["z"])))

    for i in range(len(bands)):
        kt[i] = bands[i][0]

        # Electrons
        for j in range(num_e):
            try:
                electrons[i, j] = bands[i][1]["E_e"][j]
                Psi_e[i, j, :] = bands[i][1]["Psi_e"][j].real
            except IndexError:
                electrons[i, j] = np.inf
        # Holes
        for j in range(num_hu):
            try:
                holes_U[i, j] = bands[i][1]["E_U"][j]
                Psi_g1[i, j, :] = bands[i][1]["Psi_g1"][j].real
                Psi_g2[i, j, :] = bands[i][1]["Psi_g2"][j].real
            except IndexError:
                holes_U[i, j] = np.inf

    # Conduction bands are easy to organise... the are all the same!!
    # For the hole bands, we say it is HH if the norm of Psi_g1 at k=0 is 1, and LH otherwise
    # This sorting is not very reliable as just outside k=0 the band can be strongly (mostly) LH-like
    for i in range(num_e):
        Ee.append(electrons[:, i])
        psi_e.append(Psi_e[:, i, :])

    norm = [(np.sqrt(np.trapz(p * p, x=bands[0][1]["z"]))) for p in Psi_g1[0, :]]
    for i in range(num_hu):
        if norm[i] > 0.5:
            Ehh.append(holes_U[:, i])
            psi_hh.append(Psi_g1[:, i, :])
        else:
            Elh.append(holes_U[:, i])
            psi_lh.append(Psi_g2[:, i, :])

    # We repeat the process if we are using the non-symmetric mode
    if not symmetric:
        num_hl = len(bands[0][1]["E_L"])
        holes_L = np.zeros((len(bands), num_hl))
        Psi_g3 = np.zeros((len(bands), num_hl, len(bands[0][1]["z"])))
        Psi_g4 = np.zeros((len(bands), num_hl, len(bands[0][1]["z"])))

        for i in range(len(bands)):

            for j in range(num_hl):
                try:
                    holes_L[i, j] = bands[i][1]["E_L"][j]
                    Psi_g3[i, j, :] = bands[i][1]["Psi_g3"][j].real
                    Psi_g4[i, j, :] = bands[i][1]["Psi_g4"][j].real
                except IndexError:
                    holes_L[i, j] = np.inf

                    # Plot the wavefunctions vs kt and z. Good for debugging
                    # import matplotlib.pyplot as plt
                    # import sys
                    # K, Z = np.meshgrid(kt, bands[0][1]["z"])
                    # plt.contourf(Z, K, Psi_e[:, 0, :].T**2, 50)
                    # plt.show()
                    # sys.exit()

        norm = [(np.sqrt(np.trapz(p * p, x=bands[0][1]["z"]))) for p in Psi_g4[0, :]]
        for i in range(len(norm)):
            if norm[i] > 0.5:
                Ehh.append(holes_L[:, i])
                psi_hh.append(Psi_g4[:, i, :])
            else:
                Elh.append(holes_L[:, i])
                psi_lh.append(Psi_g3[:, i, :])

        Ehh = np.array(Ehh)
        Elh = np.array(Elh)
        psi_hh = np.array(psi_hh)
        psi_lh = np.array(psi_lh)

    # Finally, we convert the lists containing the useful stuff into arrays
    Ee = np.array(Ee)
    Ehh = np.array(Ehh)
    Elh = np.array(Elh)
    psi_e = np.array(psi_e)
    psi_hh = np.array(psi_hh)
    psi_lh = np.array(psi_lh)

    return {"kt": kt,           # k points
            "Ee": Ee,           # Electron energy levels vs K
            "psi_e": psi_e,     # Electron wavefunctions vs X and K
            "Ehh": Ehh,         # Heavy hole energy levels vs K
            "psi_hh": psi_hh,   # Heavy hole wavefunctions vs X and K
            "Elh": Elh,         # Light hole energy levels vs K
            "psi_lh": psi_lh,   # Light hole wavefunctions vs X and K
            "symmetric": symmetric}  # If the structure is symmetric


def solve_bandstructure_QW(structure, num=10, kpoints=40, krange=5e9, quasiconfined=0.0, symmetric=False,
                           plot_bands=False):
    allk_t = np.linspace(0, krange, kpoints)
    bands = []

    z = structure["x"]
    fe = structure["Ve"]
    me = structure["me"]
    fhh = structure["Vhh"]
    flh = structure["Vlh"]
    g1 = structure["g1"]
    g2 = structure["g2"]
    g3 = structure["g3"]

    # First we solve at kt = 0
    new_kpoint = solve_electrons_QW_at_kt_parabolic(0.0, z, fe, me, num=(num, num), quasiconfined=quasiconfined)
    new_kpoint.update(solve_holes_QW_at_kt_4x4(0.0, z, fhh, flh, g1, g2, g3, num=(num, num), symmetric=symmetric,
                                               quasiconfined=quasiconfined))
    bands.append([0.0, new_kpoint])

    # We search only for the solutions corresponding to energy levels confined at k=0
    num_e = len(new_kpoint['E_e'])
    num_h = len(new_kpoint['E_U'])

    for kt in allk_t[1:]:
        new_kpoint = solve_electrons_QW_at_kt_parabolic(kt, z, fe, me, num=(num, num_e), quasiconfined=-0.1)
        new_kpoint.update(solve_holes_QW_at_kt_4x4(kt, z, fhh, flh, g1, g2, g3, num=(num, num_h), symmetric=symmetric,
                                                   quasiconfined=0.1))

        bands.append([kt, new_kpoint])

    print("Calculation finished!!")

    bands = band_sorting(bands, symmetric=symmetric)

    if plot_bands:
        import matplotlib.pyplot as plt

        fig, (ax1, ax2) = plt.subplots(2, 1)

        # Electrons
        ax1.plot(bands["kt"]*1e-9, bands["Ee"][0] / q, 'b', label='Ee')
        for i in range(1, len(bands["Ee"])):
            ax1.plot(bands["kt"]*1e-9, bands["Ee"][i] / q, 'b')

        # HH
        ax2.plot(bands["kt"]*1e-9, bands["Ehh"][0] / q, 'r', label='Ehh')
        for i in range(1, len(bands["Ehh"])):
            ax2.plot(bands["kt"]*1e-9, bands["Ehh"][i] / q, 'r')

        # LH
        ax2.plot(bands["kt"]*1e-9, bands["Elh"][0] / q, 'g', label='Elh')
        for i in range(1, len(bands["Elh"])):
            ax2.plot(bands["kt"]*1e-9, bands["Elh"][i] / q, 'g')

        ax1.set_ylabel('Energy (eV)')
        ax2.set_ylabel('Energy (eV)')
        plt.xlabel('k (nm$^{-1}$)')
        plt.xlim(0, max(bands["kt"]*1e-9))
        plt.setp(ax1.get_xticklabels(), visible=False)
        plt.legend()
        plt.tight_layout()
        plt.show()

    bands2 = structure.copy()
    bands2.update(bands)

    return bands2



import math  # hyperbolic functions etc in parameterisation
import os
import sys
from functools import lru_cache  # cache function calls to stop things taking forever / recalculating smae things

import numpy as np
from scipy.integrate import quad
import configparser

import solcore
from solcore.material_system import critical_point_interpolate
from solcore.parameter_system import ParameterSystem
from solcore.constants import h, c, q, kb, pi, electron_mass as m0, vacuum_permittivity
from solcore.singleton import Singleton
from solcore.absorption_calculator.sopra_db import sopra_database, compounds_info
from solcore.absorption_calculator.nk_db import nkdb_load_n, nkdb_load_k
from solcore.material_data import calculate_mobility


class MaterialSystem(metaclass=Singleton):
    """ The core class that manage the materials in solcore.
    """

    def __init__(self, sources=None):
        self.known_materials = {}
        self.sources = {}
        if sources is not None:
            self.sources = {k.lower(): sources(k) for k in sources()}

    def read(self, source, value):
        """ Reads a new source. """
        self.sources[source.lower()] = value

    def material(self, name, sopra=False, nk_db=False):
        """ This function checks if the requested material exists and creates a class
        that contains its properties, assuming that the material does not exists in
        the database, yet.

        Such class will serve as the base class for all the derived materials based on that SpecificMaterial.
        For example, N-type GaAs and P-type GaAs use the same SpecificMaterial, just with a different doping, and the
        happens with GaAs at 300K or 200K.

        The derived materials based on a SpecificMaterial are instances of the SpecificMaterial class.

        >>> GaAs = solcore.material('GaAs')      # The SpecificMaterial class
        >>> n_GaAs = GaAs(Nd=1e23)               # Instance of the class
        >>> p_GaAs = GaAs(Na=1e22)               # Another instance of GaAs with different doping

        >>> AlGaAs = solcore.material('AlGaAs')  # The SpecificMaterial class
        >>> AlGaAs_1 = AlGaAs(Al=0.3)            # Instance of the class. For compounds, the variable element MUST be present
        >>> AlGaAs_2 = AlGaAs(Al=0.7, T=290)     # Different composition and T (the default is T=300K)


        The material is created from the parameters in the parameter_system and the n and k data if available. If the
        n and k data does not exists - at all or for that composition - then n=1 and k=0 at all wavelengths. Keep in
        mind that the available n and k data is valid only at room temperature.

        :param name: Name of the material
        :param sopra: If a SOPRA material must be used, rather than the normal database material, in case both exist.
        :return: A class of that material
        """

        suffix = ''
        # First we check if the material exists. If not, some help is provided
        try:
            if sopra:
                sopra_database(Material=name)
                suffix = '_sopra'
            elif nk_db:
                suffix = '_nk'
            else:
                ParameterSystem().database.options(name)
        except configparser.NoSectionError:
            try:
                sopra_database(Material=name)
                sopra = True
                suffix = '_sopra'
            except:
                valid_materials = sorted(ParameterSystem().database.sections())
                valid_materials.remove('Immediate Calculables')
                valid_materials.remove('Final Calculables')
                print(
                    '\nMaterial ERROR: "{}" is not in the semiconductors database or in the SOPRA database. Valid semiconductor materials are: '.format(
                        name))
                for v in valid_materials:
                    if "x" in ParameterSystem().database.options(v):
                        x = ParameterSystem().database.get(v, 'x')
                        print('\t {}  \tx = {}'.format(v, x))
                    else:
                        print('\t {}'.format(v))
                print(
                    '\nIn compounds, check that the order of the elements is the correct one (eg. GaInSb is OK but InGaSb'
                    ' is not).')
                val = input('\nDo you want to see the list of available SOPRA materials (y/n)?')
                if val in 'Yy':
                    sopra_database.material_list()

                sys.exit()
        except solcore.absorption_calculator.sopra_db.SOPRAError:
            pass

        # Then we check if the material has already been created. If not, we create it.
        if name + suffix in self.known_materials:
            return self.known_materials[name + suffix]
        elif sopra:
            return self.sopra_material(name)
        elif nk_db:
            return self.nk_material(name)
        else:
            return self.parameterised_material(name)

    def parameterised_material(self, name):
        """ The function that actually creates the material class. """

        if "x" in ParameterSystem().database.options(name):
            self.composition = [ParameterSystem().database.get(name, "x")]
        else:
            self.composition = []

        class SpecificMaterial(BaseMaterial):
            material_string = name
            composition = self.composition

            def __init__(self, T=300, **kwargs):
                BaseMaterial.__init__(self, T=T, **kwargs)

        if name.lower() in self.sources.keys():
            SpecificMaterial.material_directory = self.sources[name.lower()]
            extension = ''
            if len(SpecificMaterial.composition) == 0:
                extension = '.txt'

            SpecificMaterial.n_path = os.path.join(SpecificMaterial.material_directory, 'n' + extension)
            SpecificMaterial.k_path = os.path.join(SpecificMaterial.material_directory, 'k' + extension)

        SpecificMaterial.__name__ = name

        self.known_materials[name] = SpecificMaterial
        return SpecificMaterial

    def sopra_material(self, name):
        """ Creates an optical material fromt he SOPRA database. """

        try:
            self.composition = [compounds_info.get(name, "x")]
        except:
            self.composition = []

        class SpecificMaterial(BaseMaterial):
            material_string = name
            composition = self.composition
            data = sopra_database(Material=name)

            def __init__(self, T=300, **kwargs):
                BaseMaterial.__init__(self, T=T, **kwargs)

            def __getattr__(self, attrname):  # only used for unknown attributes
                if attrname == "n":
                    return self.n_interpolated
                if attrname == "k":
                    return self.k_interpolated

                raise AttributeError(
                    'Parameter "{}" not available for this SOPRA material "{}".'.format(attrname, self.material_string))

            @lru_cache(maxsize=1)
            def load_n_data(self):
                if len(self.composition) == 0:
                    wl, n = self.data.load_n()
                    self.n_data = np.vstack((wl * 1e-9, n))

            @lru_cache(maxsize=1)
            def load_k_data(self):
                if len(self.composition) == 0:
                    wl, k = self.data.load_k()
                    self.k_data = np.vstack((wl * 1e-9, k))

            def n_interpolated(self, x):
                assert len(self.composition) <= 1, "Can't interpolate 2d spectra yet"

                try:
                    self.load_n_data()
                except:
                    print('Material "{}" does not have n-data defined. Returning "ones"'.format(self.material_string))
                    return np.ones_like(x)

                if len(self.composition) == 0:
                    y = np.interp(x, self.n_data[0], self.n_data[1])
                else:
                    wl, y, trash = self.data.load_composition(Lambda=x * 1e9,
                                                              **{self.composition[0]: self.main_fraction * 100})

                return y

            def k_interpolated(self, x):
                assert len(self.composition) <= 1, "Can't interpolate 2d spectra yet"

                try:
                    self.load_k_data()
                except:
                    print('Material "{}" does not have k-data defined. Returning "zeroes"'.format(self.material_string))
                    return np.zeros_like(x)

                if len(self.composition) == 0:
                    y = np.interp(x, self.k_data[0], self.k_data[1])
                else:
                    wl, trash, y = self.data.load_composition(Lambda=x * 1e9,
                                                              **{self.composition[0]: self.main_fraction * 100})

                return y

        SpecificMaterial.__name__ = name + '_sopra'

        self.known_materials[name + '_sopra'] = SpecificMaterial
        return SpecificMaterial

    def nk_material(self, name):
        """ Creates an optical material from the refractiveindex.info database. """

        class SpecificMaterial(BaseMaterial):

            def __init__(self, T=300, **kwargs):
                BaseMaterial.__init__(self, T=T, **kwargs)

            def __getattr__(self, attrname):  # only used for unknown attributes
                if attrname == "n":
                    return self.n_interpolated
                if attrname == "k":
                    return self.k_interpolated

                raise AttributeError(
                    'Parameter "{}" not available for this refractiveindex.info material "{}".'.format(attrname, self.material_string))

            @lru_cache(maxsize=1)
            def load_n_data(self):
                    wl, n = nkdb_load_n(name)
                    self.n_data = np.vstack((wl * 1e-6, n))  # wavelengths in DB are in microns

            @lru_cache(maxsize=1)
            def load_k_data(self):
                    wl, k = nkdb_load_k(name)
                    self.k_data = np.vstack((wl * 1e-6, k))

            def n_interpolated(self, x):
                assert len(self.composition) <= 1, "Can't interpolate 2d spectra yet"

                try:
                    self.load_n_data()
                except:
                    print('Material "{}" does not have n-data defined. Returning "ones"'.format(self.material_string))
                    return np.ones_like(x)

                y = np.interp(x, self.n_data[0], self.n_data[1])

                return y

            def k_interpolated(self, x):
                assert len(self.composition) <= 1, "Can't interpolate 2d spectra yet"

                try:
                    self.load_k_data()
                except:
                    print('Material "{}" does not have k-data defined. Returning "zeroes"'.format(self.material_string))
                    return np.zeros_like(x)

                y = np.interp(x, self.k_data[0], self.k_data[1])

                return y

        SpecificMaterial.__name__ = name + '_nk'

        self.known_materials[name + '_nk'] = SpecificMaterial
        return SpecificMaterial


class BaseMaterial:
    """ The solcore base material class
    """
    material_string = "Unnamed"
    composition = []
    main_fraction = 0
    material_directory = None
    k_path = None
    n_path = None
    strained = False

    def __init__(self, T, **kwargs):
        self.Na = kwargs["Na"] if "Na" in kwargs else 1
        self.Nd = kwargs["Nd"] if "Nd" in kwargs else 1
        self.__dict__.update(kwargs)
        self.T = T

        for element in self.composition:
            setattr(self, element, kwargs[element])

        if len(self.composition) != 0:
            self.main_fraction = kwargs[self.composition[0]]

        self.key_parameters = sorted(list(kwargs.keys()))

    def __getattr__(self, attrname):  # only used for unknown attributes.
        if attrname == "n":
            try:
                return self.__getattribute__(attrname)
            except AttributeError:
                return self.n_interpolated
        if attrname == "k":
            try:
                return self.__getattribute__(attrname)
            except AttributeError:
                return self.k_interpolated
        if attrname == "electron_affinity":
            try:
                return ParameterSystem().get_parameter(self.material_string, attrname)
            except:
                # from http://en.wikipedia.org/wiki/Anderson's_rule and GaAs values
                return (0.17 + 4.59) * q - self.valence_band_offset - self.band_gap
        if attrname == "electron_mobility":
            try:
                return ParameterSystem().get_parameter(self.material_string, attrname)
            except:
                return calculate_mobility(self.material_string, False, self.Nd, self.main_fraction)
        if attrname == "hole_mobility":
            try:
                return ParameterSystem().get_parameter(self.material_string, attrname)
            except:
                return calculate_mobility(self.material_string, True, self.Na, self.main_fraction)
        if attrname == "Nc":
            # return 2 * (2 * pi * self.eff_mass_electron_Gamma * m0 * kb * self.T / h ** 2) ** 1.5
            # Changed by Diego 22/10/18: some materials dont have _Gamma but all have the normal electron effective mass
            return 2 * (2 * pi * self.eff_mass_electron * m0 * kb * self.T / h ** 2) ** 1.5
        if attrname == "Nv":
            # Strictly speaking, this is valid only for III-V, zinc-blend semiconductors
            mhh = self.eff_mass_hh_z
            mlh = self.eff_mass_lh_z
            Nvhh = 2 * (2 * pi * mhh * m0 * kb * self.T / h ** 2) ** 1.5
            Nvlh = 2 * (2 * pi * mlh * m0 * kb * self.T / h ** 2) ** 1.5
            return Nvhh + Nvlh
        if attrname == "ni":
            return np.sqrt(self.Nc * self.Nv * np.exp(-self.band_gap / (kb * self.T)))
        if attrname == "radiative_recombination":
            inter = lambda E: self.n(E) ** 2 * self.alphaE(E) * np.exp(-E / (kb * self.T)) * E ** 2
            upper = self.band_gap + 10 * kb * self.T
            return 1.0 / self.ni ** 2 * 2 * pi / (h ** 3 * c ** 2) * quad(inter, 0, upper)[0]
        if attrname == "permittivity":
            return self.relative_permittivity * vacuum_permittivity

        kwargs = {element: getattr(self, element) for element in self.composition}
        kwargs["T"] = self.T
        return ParameterSystem().get_parameter(self.material_string, attrname, **kwargs)

    @lru_cache(maxsize=1)
    def load_n_data(self):
        if len(self.composition) == 0:
            self.n_data = np.loadtxt(self.n_path, unpack=True)
        else:
            self.n_data, self.n_critical_points = critical_point_interpolate.load_data_from_directory(self.n_path)

    @lru_cache(maxsize=1)
    def load_k_data(self):
        if len(self.composition) == 0:
            self.k_data = np.loadtxt(self.k_path, unpack=True)
        else:
            self.k_data, self.k_critical_points = critical_point_interpolate.load_data_from_directory(self.k_path)

    def n_interpolated(self, x):
        assert len(self.composition) <= 1, "Can't interpolate 2d spectra yet"

        try:
            self.load_n_data()
        except:
            print('Material "{}" does not have n-data defined. Returning "ones"'.format(self.material_string))
            return np.ones_like(x)

        if len(self.composition) == 0:
            y = np.interp(x, self.n_data[0], self.n_data[1])
        else:
            x, y, self.critical_points_n = critical_point_interpolate.critical_point_interpolate(
                self.n_data,
                self.n_critical_points,
                self.main_fraction,
                x
            )

        return y

    def k_interpolated(self, x):
        assert len(self.composition) <= 1, "Can't interpolate 2d spectra yet"

        try:
            self.load_k_data()
        except:
            print('Material "{}" does not have k-data defined. Returning "zeros"'.format(self.material_string))
            return np.zeros_like(x)

        if len(self.composition) == 0:
            y = np.interp(x, self.k_data[0], self.k_data[1])
        else:
            x, y, self.critical_points_k = critical_point_interpolate.critical_point_interpolate(
                self.k_data,
                self.k_critical_points,
                self.main_fraction,
                x
            )
        return y

    def get(self, parameter):
        return getattr(self, parameter)

    def alpha(self, wavelength):
        return 4 * math.pi * self.k(wavelength) / wavelength

    def alphaE(self, energy):
        return self.alpha(h * c / energy)

    def latex_string(self):
        s = self.material_string
        for element in self.composition:
            if self.__dict__[element] != 0:
                s = s.replace(element, "{}_{{{:.3f}}}").format(element, self.__dict__[element])
            else:
                s = s.replace(element, "")
        return "$\mathrm{{{}}}$".format(s)

    def plain_string(self):
        s = self.material_string
        for element in self.composition:
            if self.__dict__[element] != 0:
                s = s.replace(element, "{}{:.3f}").format(element, self.__dict__[element])
            else:
                s = s.replace(element, "")
        return "{}".format(s)

    def html_string(self):
        s = self.material_string
        for element in self.composition:
            if self.__dict__[element] != 0:
                s = s.replace(element, "{}<sub>{:.3f}</sub>").format(element, self.__dict__[element])
            else:
                s = s.replace(element, "")
        return "{}".format(s)

    def __repr__(self):
        parameters = ["{}={}".format(key, getattr(self, key)) for key in self.key_parameters]
        parameters_string = " ".join(parameters)
        return "<'{}' material{}>".format(
            self.material_string,
            "" if len(self.key_parameters) == 0 else " " + parameters_string)



# to add a new material to the database, so that it can be created like a built-in
# Solcore material, its path has to be added to the config file and MANIFEST.in
# Have to copy the relevant n, k and parameter files into that folder.

import os

from shutil import copyfile, move
from re import sub
from solcore import config, SOLCORE_ROOT
from solcore.parameter_system import ParameterSystem
from configparser import ConfigParser
from solcore.material_system import MaterialSystem


def create_new_material(
    mat_name, n_source, k_source, parameter_source=None, overwrite=False
):
    """
    This function adds a new material to Solcore's material_data folder, so that it can be called like a
    built-in material. It needs a name for the new material, and source files for the n and k data and other
    parameters which will be copied into the material_data/Custom folder.

    :param mat_name: the name of the new material
    :param n_source: path of the n values (txt file, first column wavelength in m, second column n)
    :param k_source: path of the n values (txt file, first column wavelength in m, second column k)
    :param: parameter_source: file with list of parameters for the new material
    """

    PARAMETER_PATH = os.path.join(config.user_folder, "custom_parameters.txt")
    if "custom" not in config.parameters():
        if not os.path.isfile(PARAMETER_PATH):
            open(PARAMETER_PATH, "a").close()
        config["Parameters", "custom"] = PARAMETER_PATH

    CUSTOM_PATH = os.path.join(config.user_folder, "custom_materials")

    # check if there is already a material with this name
    if (
        mat_name in sorted(ParameterSystem().database.sections())
        or mat_name in config.materials()
    ) and not overwrite:
        answer = input(
            f"A material named {mat_name} already exists in the database."
            f"Do you want to overwrite it [y/n]?"
        )
        if answer.lower() != "y":
            return

    # create a folder in the custom materials folders
    folder = os.path.join(CUSTOM_PATH, mat_name + "-Material")
    if not os.path.exists(folder) and folder != "":
        os.makedirs(folder)

    # copy n and k data files to the material's folder
    copyfile(n_source, os.path.join(folder, "n.txt"))
    copyfile(k_source, os.path.join(folder, "k.txt"))

    config["Materials", mat_name] = folder

    # append the parameters for the new material
    params = ConfigParser()
    params.optionxform = str
    if parameter_source is not None:
        params.read([PARAMETER_PATH, parameter_source])
        with open(PARAMETER_PATH, "w") as fp:
            params.write(fp)
    else:
        params.read([PARAMETER_PATH])
        params[mat_name] = {}
        with open(PARAMETER_PATH, "w") as fp:
            params.write(fp)
        print("Material created with optical constants n and k only.")

    ParameterSystem().read()



"""Quantum mechanics tests"""
from pytest import approx

import solcore
from solcore import material, si
from solcore.constants import vacuum_permittivity, q
from solcore.structure import Layer, Structure
import solcore.quantum_mechanics as QM
import numpy as np

# Energies in meV
my_energies = {
    "Elh": np.array([-770.0, -802.0, -803.0, -810.0, -812.0, -818.0, -823.0, -827.0]),
    "Ee": np.array([577.0, 629.0, 630.0, 640.0, 643.0, 648.0, 658.0, 663.0]),
    "Ehh": np.array(
        [
            -677.0,
            -731.0,
            -801.0,
            -801.0,
            -803.0,
            -803.0,
            -806.0,
            -807.0,
            -811.0,
            -813.0,
            -815.0,
            -816.0,
            -820.0,
            -824.0,
            -828.0,
            -832.0,
            -838.0,
            -841.0,
            -845.0,
            -849.0,
        ]
    ),
}

my_absorption = [
    1.267056856187291,
    988133.9170541381,
]  # Energy in meV and absorption coeficent in m-1


def test_kp_bands():
    """ Testing QM.kp_bands
    """
    GaAs = solcore.material("GaAs")(T=300)
    GaAsP = solcore.material("GaAsP")(P=0.3, T=300)
    InGaAs = solcore.material("InGaAs")(In=0.2, T=300)

    bands_GaAsP = QM.kp_bands(
        GaAsP,
        GaAs,
        graph=False,
        fit_effective_mass=True,
        effective_mass_direction="X",
        return_so=True,
    )
    bands_InGaAs = QM.kp_bands(
        InGaAs,
        GaAs,
        graph=False,
        fit_effective_mass=True,
        effective_mass_direction="X",
        return_so=True,
    )

    expected_bands_GaAsP = (
        1.2168382480631407e-19,
        -1.5452519153253004e-19,
        -1.4042149045828435e-19,
        -1.9192138182935611e-19,
        8.0093555784846597e-32,
        1.2472835955929216e-31,
        2.6423749777877535e-31,
        1.2393634061184521e-31,
    )
    expected_bands_InGaAs = (
        8.6764014773233634e-20,
        -1.0573103504669798e-19,
        -1.1984351916698905e-19,
        -1.6993543036257329e-19,
        6.6193386922591731e-32,
        1.3576713980579555e-31,
        8.0904387208083259e-32,
        1.2268187169919973e-31,
    )

    for i in range(len(bands_GaAsP)):
        assert bands_GaAsP[i] == approx(expected_bands_GaAsP[i])
        assert bands_InGaAs[i] == approx(expected_bands_InGaAs[i])


def test_KPbands():
    """ Testing QM.KPbands and QM.fit_effective_masses
    """
    GaAs = solcore.material("GaAs")(T=300)
    GaAsP = solcore.material("GaAsP")(P=0.3, T=300)
    InGaAs = solcore.material("InGaAs")(In=0.2, T=300)

    edges_GaAsP = QM.KPbands(GaAsP, GaAs, fraction=0.2, return_edges_only=True)
    bands_GaAsP = QM.KPbands(GaAsP, GaAs, fraction=0.2)
    masses_GaAsP = QM.fit_effective_masses(bands_GaAsP, GaAsP, GaAs, plot_result=False)

    edges_InGaAs = QM.KPbands(InGaAs, GaAs, fraction=0.2, return_edges_only=True)
    bands_InGaAs = QM.KPbands(InGaAs, GaAs, fraction=0.2)
    masses_InGaAs = QM.fit_effective_masses(
        bands_InGaAs, InGaAs, GaAs, plot_result=False
    )

    expected_edges_GaAsP = (
        1.2168382480631407e-19,
        -1.5452519153253004e-19,
        -1.4042149045828435e-19,
        -1.9192138182935611e-19,
    )

    expected_edges_InGaAs = (
        8.6764014773233634e-20,
        -1.0573103504669798e-19,
        -1.1984351916698905e-19,
        -1.6993543036257329e-19,
    )

    expected_masses_GaAsP = (
        8.049577422084102e-32,
        1.2627430248682043e-31,
        2.6577242586172804e-31,
        1.2305748108835472e-31,
    )

    expected_masses_InGaAs = (
        6.6895885457875e-32,
        1.3994390560400583e-31,
        8.142667105522975e-32,
        1.2060355194525871e-31,
    )

    for i in range(len(edges_GaAsP)):
        assert edges_GaAsP[i] == approx(expected_edges_GaAsP[i])
        assert edges_InGaAs[i] == approx(expected_edges_InGaAs[i])
        assert masses_GaAsP[i] == approx(expected_masses_GaAsP[i])
        assert masses_InGaAs[i] == approx(expected_masses_InGaAs[i])


def test_kp8x8_bulk():
    """ Testing QM.kp8x8_bulk
    """
    GaAs = solcore.material("GaAs")(T=300)
    GaAsP = solcore.material("GaAsP")(P=0.3, T=300)
    InGaAs = solcore.material("InGaAs")(In=0.2, T=300)

    bands_GaAsP = QM.kp8x8_bulk(GaAsP, GaAs)
    bands_InGaAs = QM.kp8x8_bulk(InGaAs, GaAs)

    expected_bands_GaAsP = (
        1.2168382480631407e-19,
        -1.5452519153253004e-19,
        -1.4042149045828435e-19,
        -1.9192138182935611e-19,
        8.138445281947437e-32,
        1.3543674202428507e-31,
        2.848952594319033e-31,
        1.145125159048442e-31,
    )

    expected_bands_InGaAs = (
        8.6764014773233634e-20,
        -1.0573103504669798e-19,
        -1.1984351916698905e-19,
        -1.6993543036257329e-19,
        6.77957483053393e-32,
        1.6988821114765817e-31,
        7.820551493038613e-32,
        1.1461138067300424e-31,
    )

    for i in range(len(bands_GaAsP)):
        assert bands_GaAsP[i] == approx(expected_bands_GaAsP[i])
        assert bands_InGaAs[i] == approx(expected_bands_InGaAs[i])


def test_quantum_mechanics_schrodinger():
    """ Testing schrodinger equation solver
    """
    bulk = material("GaAs")(T=293)
    barrier = material("GaAsP")(T=293, P=0.1)

    bulk.strained = False
    barrier.strained = True

    top_layer = Layer(width=si("30nm"), material=bulk)
    inter = Layer(width=si("3nm"), material=bulk)
    barrier_layer = Layer(width=si("15nm"), material=barrier)
    bottom_layer = top_layer

    E = np.linspace(1.15, 1.5, 300) * q
    alfas = np.zeros((len(E), 6))

    alfas[:, 0] = E / q

    alpha_params = {
        "well_width": si("7.2nm"),
        "theta": 0,
        "eps": 12.9 * vacuum_permittivity,
        "espace": E,
        "hwhm": si("6meV"),
        "dimensionality": 0.16,
        "line_shape": "Gauss",
    }

    QW = material("InGaAs")(T=293, In=0.2)
    QW.strained = True
    well_layer = Layer(width=si("7.2nm"), material=QW)

    my_structure = Structure(
        [
            top_layer,
            barrier_layer,
            inter,
            well_layer,
            inter,
            barrier_layer,
            inter,
            bottom_layer,
        ],
        substrate=bulk,
    )

    band_edge, bands = QM.schrodinger(
        my_structure,
        quasiconfined=0,
        num_eigenvalues=20,
        alpha_params=alpha_params,
        calculate_absorption=True,
    )

    for key in band_edge["E"]:
        for i in range(len(band_edge["E"][key])):
            band_edge["E"][key][i] = solcore.asUnit(band_edge["E"][key][i], "eV") * 1000
            band_edge["E"][key][i] = round(band_edge["E"][key][i])

    Ehh = np.all(np.equal(band_edge["E"]["Ehh"], my_energies["Ehh"]))
    Elh = np.all(np.equal(band_edge["E"]["Elh"], my_energies["Elh"]))
    Ee = np.all(np.equal(band_edge["E"]["Ee"], my_energies["Ee"]))

    idx = 100
    out = [band_edge["alpha"][0][idx] / q, band_edge["alpha"][1][idx]]

    # Test over the energies
    assert Ehh and Elh and Ee

    # Test over the absorption coefficent at a given energy
    for i, data in enumerate(out):
        assert out[i] == approx(my_absorption[i], rel=1e-4)



import numpy as np

from solcore.constants import kb, q, hbar, c
from solcore.structure import Junction
from scipy.optimize import root

from .detailed_balance import iv_detailed_balance


def iv_2diode(junction, options):
    """ Calculates the IV curve of a junction object using the 2-diode equation. All parameters needed for the calculation need to be included in the junction object. Series resistance is included at solar cell level, not at junction level. The junction is then updated with an "iv" function that calculates the IV curve at any voltage.

    :param junction: A junction object.
    :param options: Solver options.
    :return: None.
    """

    T = options.T
    light = options.light_iv
    junction.voltage = options.internal_voltages
    wl = options.wavelength

    # We get some of the minimum parameters
    R_shunt = min(junction.R_shunt, 1e14) if hasattr(junction, 'R_shunt') else 1e14
    n1 = junction.n1 if hasattr(junction, 'n1') else 1
    n2 = junction.n2 if hasattr(junction, 'n2') else 2

    try:
        # We check if we are using a radiative recombination and reference current
        # If that is the case, we solve the properties first using the DB model with the Boltzmann aproximation
        if hasattr(junction, 'reff') and hasattr(junction, 'jref'):
            options.db_mode = 'boltzmann'
            iv_detailed_balance(junction, options)

            reff = max(junction.reff, 1e-10)
            jref = junction.jref
            j01 = junction.j01

            # The reference voltage becomes
            Vref = n1 * kb * T / q * np.log(reff * jref / j01 + 1)

            # Now we can calculate the j02
            junction.j02 = ((1 - reff) * jref - Vref / R_shunt) / (np.exp(q * Vref / (n2 * kb * T)) - 1)
            j02 = junction.j02

        # If not, we are in the normal 2D equation case
        else:
            j01 = junction.j01
            j02 = junction.j02 if hasattr(junction, 'j02') else 0

            # If the saturation currents correspond to a different temperature, we update them for the current temperature.
            if hasattr(junction, 'Tref') and (T != junction.Tref):
                assert hasattr(junction,
                               'Eg'), 'ERROR: The bandgap for each junction (Eg) must be provided if the working ' \
                                      'temperature (T) is different that the reference temperature (Tref). '

                Eg = junction.Eg
                Tref = junction.Tref
                kB = kb / q
                j01 = j01 * (T / Tref) ** 3 * np.exp(-Eg / (n1 * kB) * (1 / T - 1 / Tref))
                j02 = j02 * (T / Tref) ** (5. / 3.) * np.exp(-Eg / (n2 * kB) * (1 / T - 1 / Tref))

        if not light:
            jsc = 0
        else:
            if hasattr(junction, 'jsc'):
                jsc = junction.jsc
            elif hasattr(junction, 'eqe'):
                eqe = junction.eqe
                wl, ph = options.light_source.spectrum(x=wl, output_units='photon_flux_per_m')
                jsc = q * np.trapz(eqe(wl) * ph, wl)
            else:
                jsc = 0

    except AttributeError as err:
        raise AttributeError('ERROR in 2-diode equation. Junction is missing one essential argument. {}'.format(err))

    def iv(v):
        out = j01 * (np.exp(q * v / (n1 * kb * T)) - 1) + j02 * (np.exp(q * v / (n2 * kb * T)) - 1) + v / R_shunt - jsc
        return np.minimum(out, 1e8)

    junction.jsc = jsc
    junction.current = iv(junction.voltage)
    junction.iv = iv


##########################
## Old (v4) MJ calculator.
#
# def multijunctionIV(solar_cell, V=None, photon_recycling=False, n1=1, n2=2, fraction_coupled=None,
#                     coupling_iterations=10):
#     """ Calculates the overall IV characteristics of any number of junctions numerically at the requested voltage
#     points using the 2-diodes model. If photocurrent_np is not provided, the resultant IV characteristics are purely
#     recombination currents, otherwise light IVs are returned.
#
#     If the state_object includes a reference temperature "Tref", then the reverse saturation currents are assumed to be
#     at that temperature and are therefore re-calculated at the target "T". This is only valid for small temperature
#     differences.
#
#     Conventions followed:
#         - SI Units
#         - Photocurrents: Positive.
#         - Dark Currents: Negative
#
#     :param solar_cell: Structure object containing
#
#         - T: overall temperature
#         - Tref: (optional, in case the j01 and j02 correspond to a different temperature that the target temperature)
#         - One or more Junction objects, each with (at least) the following
#         parameters in SI units:
#
#             - j01
#             - j02
#             - n1
#             - n2
#             - R_series
#             - R_shunt
#             - photocurrent (optional, is assumed = 0 if not present)
#             - Eg (optional, bandgap, only used if Tref is given)
#
#     :param photon_recycling: If photon recycling calculation must be included (NOT IMPLEMENTED YET)
#     :param V: Voltages at which to have the resulting IV curve
#     :param n1: Default n1 ideality factor (=1) if not included in the Junction objects
#     :param n2: Default n2 ideality factor (=2) if not included in the Junction objects
#     :param fraction_coupled: fraction of the emitted light coupled from one junction to the next (NOT IMPLEMENTED YET)
#     :param coupling_iterations: the maximum number of iterations during the coupling calculation (NOT IMPLEMENTED YET)
#     :return: Dictionary including:
#
#         - "IV": (V, I) Calculated IV characteristics
#         - "junction IV": [(V junc 1, I junc 1), (V junc 2, I junc 2), ...]
#         - "Rseries IV": (V, I) Calculated IV characteristics of the series resistance
#         - "V": Device voltages
#         - "I": Device currents (the same for all junctions).
#         - "Isc", Voc", "P" and "FF": In case of calculation under illumination.
#     """
#
#     # First we have to prepare the arrays necessary for the calculation
#     junctions = [obj for obj in solar_cell if type(obj) == Junction]
#
#     if hasattr(solar_cell, 'T'):
#         T = solar_cell.T
#     else:
#         T = 298
#
#     # If the saturation currents correspond to a different temperature, we update them for the current temperature.
#     if hasattr(solar_cell, 'Tref'):
#         junctions = update_j0(junctions, T, solar_cell.Tref)
#
#     J01_array = [junc.j01 for junc in junctions]
#     J02_array = [junc.j02 for junc in junctions]
#     n1_array = [junc.n1 if "n1" in junc.__dict__.keys() else n1 for junc in junctions]
#     n2_array = [junc.n2 if "n2" in junc.__dict__.keys() else n2 for junc in junctions]
#     R_shunt_array = [junc.R_shunt for junc in junctions]
#     R_series = sum([junc.R_series for junc in junctions])
#
#     # Preparation in case there is photon recycling
#     try:
#         assert photon_recycling
#         materials = [junc.material for junc in junctions]
#         qes = [junc.qe for junc in junctions]
#         e = solar_cell.energies
#     except AssertionError:
#         materials = None
#         qes = None
#         e = None
#         if photon_recycling:
#             print('WARNING: The material and QE for each junction must be provided in order to include photon '
#                   'recycling effects. Switching-off the photon recycling calculation.')
#             photon_recycling = False
#
#     # Preparation of the photocurrent of each junctions, if pressent
#     try:
#         photocurrent_array = [junc.photocurrent for junc in junctions]
#     except:
#         photocurrent_array = np.zeros_like(J01_array)  # photocurrent=0, ie dark current
#
#     # If there is no input voltage, we create a sensible set of voltages
#     if V is None:
#         output_V = np.linspace(-2, 5, 1000)
#     else:
#         output_V = V
#
#     # And now we perform the actual calculation
#     result = _multijunctionIV(J01_array, J02_array, R_shunt_array, R_series, n1_array, n2_array, T,
#                               photocurrent_array, output_V, photon_recycling, materials, qes, e, fraction_coupled,
#                               coupling_iterations)
#
#     return result
#
#
# def _multijunctionIV(J01_array, J02_array, R_shunt_array, R_series, n1_array, n2_array, T, photocurrent_array,
#                      output_V, photon_recycling=False, materials=None, qes=None, energ=None, fraction_coupled=None,
#                      coupling_iterations=10):
#     """ Calculates the overall IV characteristics of any number of junctions numerically at the requested voltage
#     points. If photocurrent is not provided, the resultant IV characteristics are purely recombination currents,
#     otherwise light IVs are returned.
#
#     Conventions followed:
#         - SI Units
#         - Photocurrents: Positive.
#         - Dark Currents: Negative
#
#     :param J01_array: Array of J01 saturation currents
#     :param J02_array: Array of J02 saturation currents
#     :param R_shunt_array: Array of shunt resistances
#     :param R_series: Combined series resistance of all juncitons
#     :param n1_array: Array of n1 ideality factors
#     :param n2_array: Array of n2 ideality factors
#     :param T: Cell temperature
#     :param photocurrent_array: Array of photocurents
#     :param photon_recycling: (False) if calculation should include photon recycling
#     :param output_V: Array of voltages in which to calculate the data
#     :param materials: Array of materials of all the junctions
#     :param qes: Array of the quantum efficiencies of all the junctions
#     :param energ: Energies of the qes
#     :param fraction_coupled: Array with the fraction of the emitted light coupled from one junction to the next.
#     :param coupling_iterations: the maximum number of iterations during the coupling calculation
#     :return: dictionary with several entries:
#         "IV": (V, I) Calculated IV characteristics
#         "junction IV": [(V junc 1, I junc 1), (V junc 2, I junc 2), ...]
#         "Rseries IV": (V, I) Calculated IV characteristics of the series resistance
#         "V": Device voltages
#         "I": Device currents (the same for all junctions).
#         "Isc", Voc", "P" and "FF": In case of calculation under illumination.
#     """
#
#     # If there is no series resistance, we make it a really small number, rather than zero, for simplicity.
#     Rs = max(np.sqrt(np.finfo(float).eps), R_series)
#
#     # The current and junction voltage arrays
#     num_jun = len(J01_array)
#     V_junction_array = np.zeros((len(output_V), len(J01_array)))
#     output_J = np.zeros_like(output_V)
#
#     # The initial guess for the numerical calculation
#     vals = np.arange(num_jun, 0.5, -1)
#     total = np.sum(vals)
#     guess = output_V[0] * vals / total
#
#     # This is the function we want to minimize: the node equations
#     def F(Vext, voltages):
#
#         # Create the output
#         output = np.zeros_like(voltages)
#
#         # First equation
#         I1 = _current(voltages[0], J01_array[0], J02_array[0], R_shunt_array[0], n1_array[0], n2_array[0],
#                       photocurrent_array[0], T)
#         output[0] = (Vext - np.sum(voltages)) / Rs - I1
#
#         # The rest of the equations
#         for j in range(1, num_jun):
#             I2 = _current(voltages[j], J01_array[j], J02_array[j], R_shunt_array[j], n1_array[j], n2_array[j],
#                           photocurrent_array[j], T)
#             output[j] = I1 - I2
#
#             I1 = I2
#
#         return output
#
#     # Now we loop over all the voltages
#     for i, VV in enumerate(output_V):
#         # We particularise the function to minimize for the current output voltage
#         def FF(voltages):
#             return F(VV, voltages)
#
#         # Solve the problem
#         result = root(FF, guess, method='lm', tol=1e-10)
#         V_junction_array[i] = result['x']
#
#         # Calculate the current and create the next initial guess
#         output_J[i] = _current(V_junction_array[i, 0], J01_array[0], J02_array[0], R_shunt_array[0], n1_array[0],
#                                n2_array[0], photocurrent_array[0], T)
#         guess = V_junction_array[i]
#
#     # Finally, we calculate the solar cell parameters
#     Isc = None
#     Voc = None
#     Pmpp = None
#     Vmpp = None
#     Impp = None
#     FF = None
#
#     # If we are calculating the light IV, we also calculate the main parameters: Jsc, Voc, FF, MPP...
#     if min(photocurrent_array) > 0:
#         Isc = -np.interp([0], output_V, output_J)[0]
#         Voc = np.interp([0], output_J, output_V)[0]
#         Power = abs(output_J[output_V < Voc] * output_V[output_V < Voc])
#         Pmpp = np.max(Power)
#         idx = np.where(Power == Pmpp)
#         Vmpp = output_V[output_V < Voc][idx]
#         Impp = output_J[output_V < Voc][idx]
#         FF = Pmpp / (Isc * Voc)
#
#     return {
#         "IV": (output_V, -output_J),
#         "V": output_V,
#         "I": -output_J,
#         "junction IV": [(V_junction_array[:, i], -output_J) for i in range(num_jun)],
#         "Rseries IV": (output_J * R_series, -output_J),
#         "Isc": Isc,
#         "Voc": Voc,
#         "FF": FF,
#         "Pmpp": Pmpp,
#         "Vmpp": Vmpp,
#         "Impp": Impp
#     }
#
#
# def MJ_IV_calculator(solar_cell, output_V, mpp_parameters=False, photon_recycling=False, materials=None, qes=None,
#                      energy=None, fraction_coupled=None, coupling_iterations=10):
#     """ Calculates the overall IV characteristics of any number of junctions numerically at the requested voltage
#     points. If photocurrent is not provided, the resultant IV characteristics are purely recombination currents,
#     otherwise light IVs are returned.
#
#     Conventions followed:
#         - SI Units
#         - Photocurrents: Positive.
#         - Dark Currents: Negative
#
#     :param solar_cell: A solar cell object with one or more junctions. The IV of the individual junctions must have been calculated.
#     :param mpp_parameters: If Isc, Voc, FF, Vmpp, Impp and Pmpp must be calculated.
#     :param photon_recycling: (False) if calculation should include photon recycling
#     :param output_V: Array of voltages in which to calculate the data
#     :param materials: Array of materials of all the junctions
#     :param qes: Array of the quantum efficiencies of all the junctions
#     :param energy: Energies of the qes
#     :param fraction_coupled: Array with the fraction of the emitted light coupled from one junction to the next.
#     :param coupling_iterations: the maximum number of iterations during the coupling calculation
#     :return: dictionary with several entries:
#         "IV": (V, I) Calculated IV characteristics
#         "junction IV": [(V junc 1, I junc 1), (V junc 2, I junc 2), ...]
#         "Rseries IV": (V, I) Calculated IV characteristics of the series resistance
#         "V": Device voltages
#         "I": Device currents (the same for all junctions).
#         "Isc", Voc", "P" and "FF": In case of calculation under illumination.
#     """
#
#     # If there is no series resistance, we make it a really small number, rather than zero, for simplicity.
#     Rs = max(np.sqrt(np.finfo(float).eps), solar_cell.R_series)
#
#     # The current and junction voltage arrays
#     num_jun = solar_cell.junctions
#     V_junction_array = np.zeros((len(output_V), num_jun))
#     output_J = np.zeros_like(output_V)
#
#     # The initial guess for the numerical calculation
#     vals = np.arange(num_jun, 0.5, -1)
#     total = np.sum(vals)
#     guess = output_V[0] * vals / total
#
#     # This is the function we want to minimize: the node equations
#     def F(voltages, Vext):
#
#         # Create the output
#         output = np.zeros_like(voltages)
#
#         # First equation
#         I1 = solar_cell(0).iv(voltages[0])
#         output[0] = (Vext - np.sum(voltages)) / Rs - I1
#
#         # The rest of the equations
#         for j in range(1, num_jun):
#             I2 = solar_cell(j).iv(voltages[j])
#             output[j] = I1 - I2
#
#             I1 = I2
#
#         return output
#
#     # Now we loop over all the voltages
#     for i, VV in enumerate(output_V):
#         # Solve the problem
#         result = root(F, guess, args=(VV), method='lm', tol=1e-10)
#         V_junction_array[i] = result['x']
#
#         # Calculate the current and create the next initial guess
#         output_J[i] = solar_cell(0).iv(V_junction_array[i, 0])
#         guess = V_junction_array[i]
#
#     # Finally, we calculate the solar cell parameters
#     Isc = None
#     Voc = None
#     Pmpp = None
#     Vmpp = None
#     Impp = None
#     FF = None
#
#     # If we are calculating the light IV, we also calculate the main parameters: Jsc, Voc, FF, MPP...
#     if mpp_parameters:
#         try:
#             Isc = -np.interp([0], output_V, output_J)[0]
#             Voc = np.interp([0], output_J, output_V)[0]
#             Power = abs(output_J[output_V < Voc] * output_V[output_V < Voc])
#             Pmpp = np.max(Power)
#             idx = np.where(Power == Pmpp)
#             Vmpp = output_V[output_V < Voc][idx]
#             Impp = output_J[output_V < Voc][idx]
#             FF = Pmpp / (Isc * Voc)
#         except Exception as err:
#             print('Error calculating the MPP parameters: {}'.format(err))
#
#     return {
#         "IV": (output_V, -output_J),
#         "V": output_V,
#         "I": -output_J,
#         "junction IV": [(V_junction_array[:, i], -output_J) for i in range(num_jun)],
#         "Rseries IV": (output_J * Rs, -output_J),
#         "Isc": Isc,
#         "Voc": Voc,
#         "FF": FF,
#         "Pmpp": Pmpp,
#         "Vmpp": Vmpp,
#         "Impp": Impp
#     }
#
#
# def simple_gp(e, n, T, V, absorption):
#     result = n ** 2 * absorption * e ** 2 / (c ** 2 * np.exp((e - q * V) / (kb * T)) - 1) / hbar ** 2
#     return result
#
#
# def _current(V_junction, J01, J02, Rshunt, n1, n2, junction_photocurrent, Temp):
#     """ Calculates the current flowing through the solar cell using the 2-diode equation. Series resistance is included
#     externally, in the nodes equation.
#
#     :param V_junction: The voltage at which to calculate the current
#     :param J01: Reverse saturation current J01, typically the radiative component
#     :param J02: Reverse saturation current J02, tipically the non-radiative component
#     :param Rshunt: Shunt resistance
#     :param n1: Ideallity factor associated to J01, typically ~1
#     :param n2: Ideallity factor associated to J02, typically ~2
#     :param junction_photocurrent: Photocurrent
#     :param Temp: Junction temperature
#     :return: The current in the above conditions
#     """
#     output = J01 * (np.exp(q * V_junction / (n1 * kb * Temp)) - 1) + \
#              J02 * (np.exp(q * V_junction / (n2 * kb * Temp)) - 1) + \
#              V_junction / Rshunt - junction_photocurrent
#     return output
#

def calculate_J01(Eg_in_eV, T, n):
    """ Calculate the reverse saturation current J01, assumed radiative, considering an absorption equal to 1 above the
    bandgap. Light trapping is included by considering the refractive index of the material:

    .. math:: J_{01} = \\frac {q n^2 k_b T} {2 \\pi ^2 c^2 \\hbar ^3} e^{\\frac{-E_g}{k_b T}} (E_g^2 + 2 k_b T E_g + 2 k_b^2 T^2)


    :param Eg_in_eV: Bandgap  in eV
    :param T: Cell temperature
    :param n: Refractive index of the material
    :return: The reverse saturation current J01
    """
    Eg = Eg_in_eV * q
    Term1 = 2 * np.pi * n ** 2 * q / (4 * np.pi ** 3 * hbar ** 3 * c ** 2)
    Term2 = kb * T * np.exp(-Eg / (kb * T))
    Term3 = Eg ** 2 + (2 * Eg * kb * T) + (2 * kb ** 2 * T ** 2)

    J01 = Term1 * Term2 * Term3
    return J01


def calculate_J02_from_Voc(J01, Jsc, Voc, T, R_shunt=1e15):
    """ Calculates J02 based on the J01, Jsc and the Voc. It is just the result of solving the 2-diode equation for J02.
    Ideality factors n1 and n2 are assumed to be equal to 1 and 2, respectively.

    :param J01: Reverse saturation current J01, typically the radiative component
    :param Jsc: Short circuit current (=photocurrent)
    :param Voc: Open circuit voltage
    :param T: Temperature
    :param R_shunt: Shunt resistance (default = 1e15)
    :return: The reverse saturation current J02
    """
    Term1 = Jsc - J01 * (np.exp(q * Voc / (kb * T)) - 1) - Voc / R_shunt
    Term2 = np.exp(q * Voc / (2 * kb * T)) - 1

    J02 = Term1 / Term2
    return J02


def calculate_J02_from_rad_eff(J01, radiative_efficiency, V, T, R_shunt=1e15):
    """ Calculates J02 based on J01 and a radiative efficiency at a given voltage and temperature. Ideality factors n1
    and n2 are assumed to be equal to 1 and 2, respectively.

    :param J01: Reverse saturation current J01, typically the radiative component
    :param radiative_efficiency: Fraction of the dark current that is radiative
    :param V: Operating voltage
    :param T: Temperature
    :param R_shunt: Shunt resistance (default = 1e15)
    :return:
    """

    Term1 = J01 * (np.exp(q * V / (kb * T)) - 1)
    Term2 = 1 / radiative_efficiency - 1
    Term3 = np.exp(q * V / (2 * kb * T)) - 1

    J02 = (Term1 * Term2 + V / R_shunt) / Term3

    return J02


def calculate_j02_from_J01_Jsc_reference_radiative_efficiency(J01, Jsc_ref, radiative_efficiency, T):
    Voc_reference = (kb * T / q) * np.log((Jsc_ref * radiative_efficiency / J01) + 1)
    J02 = (Jsc_ref - (radiative_efficiency * Jsc_ref)) / (np.exp(q * Voc_reference / (2 * kb * T)) - 1)

    return J02


def update_j0(junctions, T, Tref):
    """ Updates the reverse saturation currents for the target temperature knowing their values at a reference
    temperature.

    :param junctions: List of junctions
    :param T: Target temperature
    :param Tref: Working temperature
    :return: List of junctions with their saturation currents updated for the new temperatures
    """
    kB = kb / q

    for junc in junctions:
        assert hasattr(junc, 'Eg'), 'ERROR: The bandgap for each junction (Eg) must be provided if the working ' \
                                    'temperature (T) is different that the reference temperature (Tref). '
        junc.j01 = junc.j01 * (T / Tref) ** 3 * np.exp(-junc.Eg / kB * (1 / T - 1 / Tref))
        junc.j02 = junc.j02 * (T / Tref) ** (5. / 3.) * np.exp(-0.5 * junc.Eg / kB * (1 / T - 1 / Tref))

    return junctions



from collections import defaultdict
import re  # tools to manage regular expresions
import numpy as np
from typing import Optional, Callable

import solcore
from solcore.source_managed_class import SourceManagedClass
from solcore.constants import *


class UnitError(Exception):
    def __init__(self, msg):
        BaseException.__init__(self, msg)


class WrongDimensionError(Exception):
    def __init__(self, msg):
        BaseException.__init__(self, msg)


def generateConversionDictForSISuffix(suffix, centi=False, deci=False, non_base_si_factor=1):
    prefixes = "Y,Z,E,P,T,G,M,k,,m,u,n,p,f,a,z,y".split(",")
    exponents = list(range(8, -9, -1))

    if centi:
        prefixes.append("c")
        exponents.append(-2. / 3.)

    if deci:
        prefixes.append("d")
        exponents.append(-1. / 3.)

    unitNames = ["%s%s" % (prefix, suffix) for prefix in prefixes]
    conversion = [1000. ** exponent * non_base_si_factor for exponent in exponents]

    return dict(zip(unitNames, conversion))


class UnitsSystem(SourceManagedClass):
    """ Contains all the functions related with the conversion of units. While defined inside this class, most of these
    functions are available outside it, being decorated with 'breakout' (see 'Singleton')"""

    def __init__(self, sources: Optional[Callable] = None):
        super().__init__({k: sources(k) for k in sources()})
        self.separate_value_and_unit_RE = re.compile(u"([-+]?[0-9]*\.?[0-9]+(?:[eE][-+]?[0-9]+)?)(?:[ \t]*(.*))?")
        self.split_units_RE = re.compile(u"(?:([^ \+\-\^\.0-9]+)[\^]?([\-\+]?[^ \-\+]*)?)")
        self.siConversions = {}
        self.dimensions = defaultdict(dict)
        self.read()

    def read(self, source=None, value=None):
        """ Reads the units file and creates a database with all units and conversion factors. """
        if source is not None:
            super().read(source, value)

        for dimension in self.database.sections():
            units = self.database.options(dimension)
            for unit in units:
                if "META_GenerateConversions" in unit:
                    expression = self.database.get(dimension, unit)
                    si_base_unit = expression.split()[0]
                    centi = "centi" in expression
                    deci = "deci" in expression
                    non_base_si_factor = self.safe_eval(expression.split()[1]) if "altbase" in expression else 1
                    dimension_conversions = generateConversionDictForSISuffix(
                        si_base_unit,
                        centi=centi,
                        deci=deci,
                        non_base_si_factor=non_base_si_factor
                    )
                    self.siConversions.update(dimension_conversions)
                    self.dimensions[dimension].update(dimension_conversions)
                    continue
                string_expression = self.database.get(dimension, unit)
                self.siConversions[unit] = self.safe_eval(string_expression)
                self.dimensions[dimension][unit] = self.siConversions[unit]

    def safe_eval(self, string_expression):
        return eval(string_expression, {"__builtins__": {}}, {"constants": solcore.constants})

    def siUnits(self, value, unit):
        """ Convert value from unit to equivalent si-unit

        >>> print(siUnits(1,"mm")) # yields meters
        0.001
        >>> print(siUnits(1,"um")) # yields meters
        1e-06

        :param value: the value to convert
        :param unit: the units of the value
        :return: the value expresed in SI units
        """
        if unit is None or value is None:
            return value

        units_list = self.split_units_RE.findall(unit)
        for unit, power in units_list:
            power = float(power) if power != '' else 1
            value = value * np.power((self.siConversions[unit]),
                                     power)  ### caution, *= is WRONG because it modifies original obj. DO NOT WANT

        return value

    def asUnit(self, value, unit):
        """ Converts from si unit to other unit. It is the reversed of siUnits function

        >>> print(asUnit(1, "mA")) # print 1A in mA.
        1000.0

        :param value: the value to convert, assumed in SI units
        :param unit: the new units
        :param dimension: the value expressed in the new units.
        :return:
        """
        if unit is None or value is None:
            return value

        units_list = self.split_units_RE.findall(unit)
        for unit, power in units_list:
            power = float(power) if power != '' else 1
            value = value / (self.siConversions[unit]) ** power  ### caution, /= is WRONG because it modifies original obj. DO NOT WANT

        return value

    def si(self, *args):
        """ Utility function that forwards to either siUnit or siUnitFromString"""
        if type(args[0]) == str:
            return self.siUnitFromString(*args)
        return self.siUnits(*args)

    def siUnitFromString(self, string):
        """ Converts a string of a number with units into si units of that quantity

        >>> print(si("5 mm s-1")) # output in m/s
        0.005
        >>> print(si("5e-0mm-2")) # output in m2
        5000000.0
        >>> print(si("5"))
        5.0

        :param string: the string to convert
        :return: the value in SI units
        """
        # if unit is None or value is None:
        #     return value

        matchObj = self.separate_value_and_unit_RE.match(string)
        value, unit = matchObj.groups()
        value = float(value)
        units_list = self.split_units_RE.findall(unit)
        for unit, power in units_list:
            power = float(power) if power != '' else 1
            value *= (self.siConversions[unit]) ** power
        return value

    def convert(self, value, from_unit, to_unit):
        """ Converts between comparable units, does NOT check if units are comparable.

        >>> print(convert(1, "nm", "mm"))
        1e-06
        >>> print(convert(1, "um", "nm"))
        1000.0
        >>> print(convert(1, "cm s-1", "km h-1"))
        0.036

        :param value: the value ot convert
        :param from_unit: the original unit
        :param to_unit: the final unit
        :return: the value expressed in the final unit
        """
        return self.asUnit(self.siUnits(value, from_unit), to_unit)

    def eVnm(self, value):
        """ Bi-directional conversion between nm and eV.

        >>> print('%.3f'%eVnm(1000))
        1.240
        >>> print('%i'%round(eVnm(1)))
        1240

        :param value: a number with units [nm] or [eV].
        :return: either the conversion [nm] --> [eV], or [eV] --> [nm]
        """
        factor = self.asUnit(h, "eV") * self.asUnit(c, "nm")
        return factor / value

    def nmJ(self, value):
        """ Bi-directional conversion between nm and J.

        >>> print(nmJ(1000))
        1.9864452126e-19
        >>> print(nmJ(2e-18))
        99.3222606298

        :param value: a number with units [nm] or [J].
        :return: either the conversion [nm] --> [J], or [J] --> [nm]
        """
        factor = h * c
        return factor / self.siUnits(value, "nm")

    def mJ(self, value):
        """ Bi-directional conversion between m and J.

        >>> print(mJ(1000))
        1.986445212595144e-25
        >>> print(mJ(2e-18))
        9.93222606297572e-08

        :param value: a number with units [m] or [J].
        :return: either the conversion [m] --> [J], or [J] --> [m]
        """
        factor = h * c
        return factor / value

    def nmHz(self, value):
        """ Bi-directional conversion between nm and Hz.

        :param value: a number with units [nm] or [Hz].
        :return: Either a number which is the conversion [nm] --> [Hz] or [Hz] --> [nm]
        """
        factor = self.asUnit(c, "nm s-1")
        return factor / value

    def spectral_conversion_nm_ev(self, x, y):
        """ Bi-directional conversion between a spectrum per nanometer and a spectrum per electronvolt.

        Example:
        1) nm --> eV conversion
        wavelength_nm
        photon_flux_per_nm
        energy_ev, photon_flux_per_ev = spectral_conversion_nm_ev(wavelength_nm, photon_flux_per_nm)

        2) eV --> nm conversion
        energy_ev
        photon_flux_per_ev
        wavelength_nm, photon_flux_per_nm = spectral_conversion_nm_ev(energy_ev, photon_flux_per_ev)

        Discussion:
        A physical quantities such as total number of photon in a spectrum or
        total energy of a spectrum should remain invariant after a transformation
        to different units. This is called a spectral conversion. This function
        is bi-directional because the mathematics of the conversion processes
        is symmetrical.

        >>> import numpy as np
        >>> x = np.array([1,2,3])
        >>> y = np.array([1,1,1])
        >>> area_before = np.trapz(y, x=x)
        >>> x_new, y_new = spectral_conversion_nm_ev(x, y)
        >>> area_after = np.trapz(y_new, x=x_new)
        >>> compare_floats(area_before, area_after, relative_precision=0.2)
        True

        :param x: abscissa of the spectrum in units of [nm] or [eV]
        :param y: ordinate of the spectrum in units of [something/nm] or [something/eV]
        :return: A tuple (x, y) which has units either [eV, something/eV] or [nm. something/nm].
        """
        x_prime = self.eVnm(x)
        conversion_constant = self.asUnit(h, "eV s") * self.asUnit(c, "nm s-1")
        y_prime = y * conversion_constant / x_prime ** 2
        y_prime = reverse(y_prime)  # Wavelength ascends as electronvolts decends therefore reverse arrays
        x_prime = reverse(x_prime)
        return (x_prime, y_prime)

    def spectral_conversion_nm_hz(self, x, y):
        """ Bi-directional conversion between a spectrum per nanometer and a spectrum per Hertz.

        Example:
        1) nm --> Hz conversion
        wavelength_nm
        photon_flux_per_nm
        energy_hz, photon_flux_per_hz = spectral_conversion_nm_hz(wavelength_nm, photon_flux_per_nm)

        2) Hz --> nm conversion
        energy_hz
        photon_flux_per_hz
        wavelength_nm, photon_flux_per_nm = spectral_conversion_nm_ev(energy_hz, photon_flux_per_hz)

        Discussion:
        A physical quantities such as total number of photon in a spectrum or
        total energy of a spectrum should remain invariant after a transformation
        to different units. This is called a spectral conversion. This function
        is bi-directional because the mathematics of the conversion processes
        is symmetrical.

        >>> import numpy as np
        >>> x = np.array([1,2,3])
        >>> y = np.array([1,1,1])
        >>> area_before = np.trapz(y, x=x)
        >>> x_new, y_new = spectral_conversion_nm_hz(x, y)
        >>> area_after = np.trapz(y_new, x=x_new)
        >>> compare_floats(area_before, area_after, relative_precision=0.2)
        True

        :param x: abscissa of the spectrum in units of [nm] or [Hz]
        :param y: ordinate of the spectrum in units of [something/nm] or [something/Hz]
        :return: A tuple (x, y) which has units either [eV, something/nm] or [nm. something/Hz].
        """
        x_prime = self.nmHz(x)
        conversion_constant = self.asUnit(c, "nm s-1")
        y_prime = y * conversion_constant / x_prime ** 2
        y_prime = reverse(y_prime)  # Wavelength ascends as frequency decends therefore reverse arrays
        x_prime = reverse(x_prime)
        return (x_prime, y_prime)

    def sensibleUnits(self, value, dimension, precision=2):
        """ Attempt to convert a physical quantity of a particular dimension to the most sensible units

        >>> print(sensibleUnits(0.001,"length",0))
        1 mm
        >>> print(sensibleUnits(1000,"length",0))
        1 km
        >>> print(sensibleUnits(si("0.141 days"),"time", 5))
        3.38400 h

        :param value: The value to re-calculate in SI units
        :param dimension: The dimension of the value. Possible values are: 'luminous intensity', 'pressure', 'time',
         'angle', 'temperature', 'current', 'force', 'charge', 'power', 'voltage', 'resistance', 'mass', 'length', 'energy'
        :param precision: Precission of the converted value.
        :return: A string with the value in the more 'sensible' units and the units.
        """
        negative = ""
        if value < 0:
            value *= -1
            negative = "-"
        formatting = "%s%%.%if %%s" % (negative, precision)
        d = self.dimensions[dimension]
        possibleUnits = d.keys()
        if value == 0:
            return formatting % (0, "")
        allValues = [abs(np.log10(self.asUnit(value, unit))) for unit in possibleUnits]
        bestUnit = possibleUnits[allValues.index(min(allValues))]
        return formatting % (self.asUnit(value, bestUnit), bestUnit)

    def eV(self, e):
        """ Transform an energy value in SI units in a string expresing the value in eV.

        >>> print(eV(1e-19))
        0.624 eV

        :param e: Energy value in SI units
        :return: A string with the energy converted in eV and its units.
        """
        return "%.3f eV" % self.asUnit(e, "eV")

    def guess_dimension(self, unit):
        """ Guess the dimension of a unit.

        >>> print guess_dimension("nm")
        length

        :param unit: the unit.
        :return: None
        """
        possibilities = [key for key in self.dimensions.keys() if unit in self.dimensions[key]]

        assert len(possibilities) != 0, "Guessing dimension of '%s': No candidates found" % unit
        assert len(
            possibilities) == 1, "Guessing dimension of '%s': Multiple candidates found, please convert manually. (%s)" % (
            unit, ", ".join(possibilities))

        return possibilities[0]

    def list_dimensions(self):
        for dim in self.dimensions.keys():
            print(
                "%s: %s" % (dim, ", ".join([k for k in self.dimensions[dim].keys() if k is not None and k is not ""])))


def compare_floats(a, b, absoulte_precision=1e-12, relative_precision=None):
    """ Returns true if the absolute difference between the numbers a and b is less than the precision.

    Arguments:
    a -- a float
    b -- a float

    Keyword Arguments (optional):
    absolute_precision -- the absolute precision, abs(a-b) of the comparison.
    relative_precision -- the relative precision, max(a,b)/min(a,b) - 1. of the comparison.

    Returns:
    True if the numbers are the same within the limits of the precision.
    False if the number are not the same within the limits of the precision.
    """

    if relative_precision is None:
        absolute = abs(a - b)
        if absolute < absoulte_precision:
            return True
        else:
            return False
    else:
        relative = max(a, b) / min(a, b) - 1.
        if relative < relative_precision:
            return True
        else:
            return False


def independent_nm_ev(x, y):
    return UnitsSystem().eVnm(x)[::-1], y[::-1]


def independent_nm_J(x, y):
    return UnitsSystem().nmJ(x)[::-1], y[::-1]


def independent_m_J(x, y):
    return reverse(UnitsSystem().mJ(x)), reverse(y)


def reverse(x):
    return x[::-1]


if __name__ == '__main__':
    UnitsSystem()

    print(solcore.eVnm(1240))



import numpy as np
from solcore.constants import kb, q


def resistive_tunnel_junction(junction, options):
    """ Calculates the IV curve of a tunnel junction when it is modelled as a simple resistor. The minimum resistance of the junction is always 1e-16.

    :param junction: A junction object.
    :param options: Solver options.
    :return: None.
    """

    def iv(v):
        return v / junction.R

    def vi(j):
        return j * junction.R

    junction.voltage = options.internal_voltages
    junction.current = iv(junction.voltage)
    junction.iv = iv
    junction.vi = vi


def parametric_tunnel_junction(junction, options):
    """ Calculates the IV curve of a tunnel junction when it is modelled using a set of empirical parameters, such as peak curent and voltage, valley current and voltage, etc. The total current of the tunnel junction is summ of 3 components: the tunnel current, the excess current and the difussion current.

    :param junction: A junction object.
    :param options: Solver options.
    :return: None.
    """

    T = options.T
    junction.voltage = options.internal_voltages

    try:
        # First we calculate the tunnel current
        jp = junction.j_peak
        vp = junction.v_peak

        def tunnel_current(v):
            return jp * v / vp * np.exp(1 - v / vp)

        junction.tunnel_current = tunnel_current

        # Now we calculate the excess current due to carrier tunnelling by way of states within the forbidden gap
        jva = junction.j_valley
        vv = junction.v_valley
        pref = junction.prefactor

        def excess_current(v):
            return jva * np.exp(pref * (v - vv))

        junction.excess_current = excess_current

        # Finally we calculate the diffusion current
        j01 = junction.j01

        def diffusion_current(v):
            return j01 * np.exp(q * v / kb / T)

        junction.diffusion_current = diffusion_current
    except AttributeError:
        raise

    # And we put everything together in an IV curve function
    def iv(v):
        return junction.tunnel_current(v) + junction.excess_current(v) + junction.diffusion_current(v)

    junction.current = iv(junction.voltage)

    # Also, we calculate the inverse, needed for the calculation of the IV curve in MJ solar cells
    MJ_current = np.where(junction.voltage <= junction.v_peak, junction.current,
                          np.maximum(iv(junction.v_peak), junction.current))

    # And the corresponding function doing that
    if junction.pn:
        def vi(j):
            return np.interp(-j, MJ_current, -junction.voltage)
    else:
        def vi(j):
            return np.interp(j, MJ_current, junction.voltage)

    junction.iv = iv
    junction.vi = vi


def external_tunnel_junction(junction, options):
    """ Calculates the IV curve of a tunnel junction when it is modelled with external data. The external voltage and current must be defined with the input parameters external_voltage and external_current, respectively. It assumes that the interesting part of the curve is in the 1st quadrant (V>0 and I>0).

    :param junction: A junction object.
    :param options: Solver options.
    :return: None.
    """

    try:
        # We put everything together in an IV curve function using the external data
        def iv(v):
            return np.interp(v, junction.external_voltage, junction.external_current)

    except AttributeError:
        raise

    # Also, we calculate the inverse, needed for the calculation of the IV curve in MJ solar cells
    MJ_current = np.zeros_like(junction.external_current)
    MJ_current[0] = junction.external_current[0]
    for i in range(len(junction.external_current) - 1):
        MJ_current[i + 1] = max(MJ_current[i], junction.current[i + 1])

    # And the corresponding function doing that
    def vi(j):
        return np.interp(j, MJ_current, junction.external_voltage)

    junction.voltage = options.internal_voltages
    junction.current = iv(junction.voltage)
    junction.iv = iv
    junction.vi = vi


def example_resistive_tunnel_junction(show=True):
    from solcore.structure import TunnelJunction
    from solcore.solar_cell_solver import default_options
    import matplotlib.pyplot as plt

    # Parametric tunnel example
    my_tunnel = TunnelJunction(R=0.05)
    resistive_tunnel_junction(my_tunnel, default_options)

    if show:
        v = my_tunnel.voltage

        plt.plot(v, my_tunnel.current, 'k', linewidth=2, label='Total')

        plt.legend(fontsize=12)
        plt.ylim(0, 1.5)
        plt.xlim(0, 1)
        plt.ylabel('Current Density(A/$m^2$)', fontsize=12)
        plt.xlabel('Voltage(V)', fontsize=12)
        plt.tick_params(labelsize=12)
        plt.tight_layout()
        plt.show()

    return my_tunnel


def example_parametric_tunnel_junction(show=True):
    from solcore.structure import TunnelJunction
    from solcore.solar_cell_solver import default_options
    import matplotlib.pyplot as plt

    # Parametric tunnel example
    my_tunnel = TunnelJunction(v_peak=0.1, j_peak=1, v_valley=0.5, j_valley=0.1, prefactor=10, j01=1e-11)
    parametric_tunnel_junction(my_tunnel, default_options)

    if show:
        v = my_tunnel.voltage

        plt.plot(v, my_tunnel.tunnel_current(v), 'r--', label='Tunnel')
        plt.plot(v, my_tunnel.excess_current(v), 'g--', label='Excess')
        plt.plot(v, my_tunnel.diffusion_current(v), 'b--', label='Diffusion')
        plt.plot(v, my_tunnel.current, 'k', linewidth=2, label='Total')

        plt.legend(fontsize=12)
        plt.ylim(0, 1.5)
        plt.xlim(0, 1)
        plt.ylabel('Current Density(A/$m^2$)', fontsize=12)
        plt.xlabel('Voltage(V)', fontsize=12)
        plt.tick_params(labelsize=12)
        plt.tight_layout()
        plt.show()

    return my_tunnel


if __name__ == '__main__':
    import matplotlib.pyplot as plt
    from solcore.structure import TunnelJunction
    from solcore.solar_cell_solver import default_options

    # tunnel = example_parametric_tunnel_junction(False)
    # example_resistive_tunnel_junction()

    tunnel = TunnelJunction(v_peak=0.2, j_peak=7.5e4, v_valley=1, j_valley=40000, prefactor=5, j01=1e-23, kind='parametric')
    parametric_tunnel_junction(tunnel, default_options)

    v = tunnel.voltage

    I = np.linspace(0, 105000, 100)

    plt.plot(v, tunnel.tunnel_current(v), 'r--', label='Tunnel')
    plt.plot(v, tunnel.excess_current(v), 'g--', label='Excess')
    plt.plot(v, tunnel.diffusion_current(v), 'b--', label='Diffusion')
    plt.plot(v, tunnel.current, 'k', linewidth=3, color='DimGray', label='Total')
    # plt.plot(v, MJ_current, 'k-o', linewidth=2, label='MJ current')
    # plt.plot(tunnel.vi(I), I, '--', color='grey', linewidth=2, label='MJ current')
    plt.plot((0.2, 0.9), (100, 10), 'ko')

    plt.annotate('V$_P$, J$_P$', xy=(0.2, 110), fontsize=12)
    plt.annotate('V$_V$, J$_V$', xy=(0.6, 10), fontsize=12)


    plt.legend(fontsize=12, frameon=False)
    plt.ylim(0, 105000)
    plt.xlim(0, 2)
    plt.ylabel('Current Density(A/$m^2$)', fontsize=12)
    plt.xlabel('Voltage(V)', fontsize=12)
    plt.tick_params(labelsize=12)
    plt.tight_layout()

    plt.show()



from setuptools import find_packages
from numpy.distutils.core import Extension, setup

import os
import sys
from configparser import ConfigParser


def gen_data_files(*dirs):
    """Creates the list of files (not necessarily python files) that need to be
    installed together with the rest of stuff"""
    results = []
    exclude = [".DS_Store", "__pycache__", "egg", ".git"]
    for src_dir in dirs:
        for root, dirs, files in os.walk(src_dir):
            if not any(sub in root for sub in exclude):
                results.append(
                    (root, [os.path.join(root, f) for f in files if f not in exclude])
                )
    return results


here = os.path.abspath(os.path.dirname(__file__))
default_config = os.path.join(here, "solcore", "solcore_config.txt")
config = ConfigParser()
config.read([default_config])

# We give the option of compiling - and installing - the extension modules
if (("--with_pdd" in sys.argv) or
        (('SOLCORE_WITH_PDD' in os.environ) and
         (os.environ['SOLCORE_WITH_PDD'] == "1"))):
    sources = os.path.join("solcore", "poisson_drift_diffusion",
                           "DDmodel-current.f95")
    ext = [
        Extension(
            name="solcore.poisson_drift_diffusion.ddModel",
            sources=[sources],
            f2py_options=["--quiet"],
            extra_link_args=["-static", "-static-libgfortran", "-static-libgcc"] if sys.platform == "win32" else None
        )
    ]
    if "--with_pdd" in sys.argv:
        sys.argv.remove("--with_pdd")
else:
    ext = []

# Option for updating the manifest
if "update_manifest" in sys.argv:
    # Update the MANIFEST.in file with all the data from within solcore
    include = "solcore"
    exclude = ["__pycache__", "egg", "darwin", "cpython"]
    with open("MANIFEST.in", "w", encoding="utf-8") as f:
        for root, dir, files in os.walk("."):
            if not any(sub in root for sub in exclude) and root[2:9] == include:
                try:
                    files.remove(".DS_Store")
                except ValueError:
                    pass
                for file in files:
                    if any(sub in file for sub in exclude):
                        continue
                    include_line = "include " + os.path.join(root[2:], file) + "\n"
                    f.write(include_line)

    sys.exit()

# Get the long description from the README file
with open(os.path.join(here, "README.md"), encoding="utf-8") as f:
    long_description = f.read()


install_requires = [
    "numpy",
    "matplotlib",
    "scipy",
    "tmm",
    "natsort",
    "regex",
    "cycler",
    "pyyaml",
    "yabox",
    "joblib",
]
tests_require = [
    "pytest",
    "pytest-cov",
    "pytest-mock",
    "nbconvert",
    "nbformat",
    "pytest-rerunfailures",
    "pytest-xdist",
]
docs_require = ["Sphinx", "recommonmark"]
extras_require = {
    "dev": tests_require + docs_require + ["pre-commit"],
    "docs": docs_require,
    "test": tests_require,
}


setup(
    name="solcore",
    version=config.get("Configuration", "version"),
    description="Python-based solar cell simulator",
    long_description=long_description,
    url="https://github.com/qpv-research-group/solcore5",
    download_url="https://github.com/qpv-research-group/solcore5/archive/v{}.tar.gz".format(
        config.get("Configuration", "version")
    ),
    project_urls={
        "Documentation": "http://solcore5.readthedocs.io",
        "Solcore research paper": "https://doi.org/10.1007/s10825-018-1171-3",
    },
    author="The Quantum Photovoltaics Group",
    author_email="d.alonso-alvarez@imperial.ac.uk",
    license="GNU LGPL",
    python_requires=">=3.7",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved",
        "License :: OSI Approved :: GNU Lesser General Public License v3 or later (LGPLv3+)",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Topic :: Scientific/Engineering",
        "Topic :: Scientific/Engineering :: Physics",
    ],
    ext_modules=ext,
    keywords="photovoltaics modelling physics",
    packages=find_packages(exclude=[]),
    package_data={"": ["*.*"]},
    data_files=gen_data_files("solcore"),
    include_package_data=True,
    setup_requires="pytest-runner",
    install_requires=install_requires,
    tests_require=tests_require,
    extras_require=extras_require,
)
